# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Dict, Union

from supertokens_python.logger import log_debug_message
from supertokens_python.recipe.passwordless.interfaces import (
    APIInterface,
    APIOptions,
    ConsumeCodeExpiredUserInputCodeError,
    ConsumeCodeIncorrectUserInputCodeError,
    ConsumeCodePostExpiredUserInputCodeError,
    ConsumeCodePostIncorrectUserInputCodeError,
    ConsumeCodePostOkResult,
    ConsumeCodePostRestartFlowError,
    ConsumeCodeRestartFlowError,
    CreateCodePostOkResult,
    CreateNewCodeForDeviceOkResult,
    CreateNewCodeForDeviceUserInputCodeAlreadyUsedError,
    EmailExistsGetOkResult,
    PasswordlessLoginEmailTemplateVars,
    PhoneNumberExistsGetOkResult,
    ResendCodePostOkResult,
    ResendCodePostRestartFlowError,
)
from supertokens_python.recipe.passwordless.types import (
    PasswordlessLoginSMSTemplateVars,
)
from supertokens_python.recipe.passwordless.utils import (
    ContactEmailOnlyConfig,
    ContactEmailOrPhoneConfig,
    ContactPhoneOnlyConfig,
)
from supertokens_python.recipe.session.asyncio import create_new_session
from supertokens_python.types import GeneralErrorResponse
from ...emailverification import EmailVerificationRecipe
from ...emailverification.interfaces import CreateEmailVerificationTokenOkResult


class APIImplementation(APIInterface):
    async def create_code_post(
        self,
        email: Union[str, None],
        phone_number: Union[str, None],
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[CreateCodePostOkResult, GeneralErrorResponse]:
        user_input_code = None
        if api_options.config.get_custom_user_input_code is not None:
            user_input_code = await api_options.config.get_custom_user_input_code(
                user_context
            )
        response = await api_options.recipe_implementation.create_code(
            email, phone_number, user_input_code, user_context
        )
        magic_link = None
        user_input_code = None
        flow_type = api_options.config.flow_type
        if flow_type in ("MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"):
            magic_link = (
                api_options.app_info.website_domain.get_as_string_dangerous()
                + api_options.app_info.website_base_path.get_as_string_dangerous()
                + "/verify"
                + "?rid="
                + api_options.recipe_id
                + "&preAuthSessionId="
                + response.pre_auth_session_id
                + "#"
                + response.link_code
            )
        if flow_type in ("USER_INPUT_CODE", "USER_INPUT_CODE_AND_MAGIC_LINK"):
            user_input_code = response.user_input_code

        if isinstance(api_options.config.contact_config, ContactEmailOnlyConfig) or (
            isinstance(api_options.config.contact_config, ContactEmailOrPhoneConfig)
            and email is not None
        ):
            if email is None:
                raise Exception("Should never come here")

            log_debug_message("Sending passwordless login email to %s", email)
            passwordless_email_delivery_input = PasswordlessLoginEmailTemplateVars(
                email=email,
                user_input_code=user_input_code,
                url_with_link_code=magic_link,
                code_life_time=response.code_life_time,
                pre_auth_session_id=response.pre_auth_session_id,
            )
            await api_options.email_delivery.ingredient_interface_impl.send_email(
                passwordless_email_delivery_input, user_context
            )
        elif isinstance(
            api_options.config.contact_config,
            (ContactEmailOrPhoneConfig, ContactPhoneOnlyConfig),
        ):
            if phone_number is None:
                raise Exception("Should never come here")
            log_debug_message("Sending passwordless login SMS to %s", phone_number)
            sms_input = PasswordlessLoginSMSTemplateVars(
                phone_number=phone_number,
                user_input_code=user_input_code,
                url_with_link_code=magic_link,
                code_life_time=response.code_life_time,
                pre_auth_session_id=response.pre_auth_session_id,
            )
            await api_options.sms_delivery.ingredient_interface_impl.send_sms(
                sms_input, user_context
            )

        return CreateCodePostOkResult(
            response.device_id, response.pre_auth_session_id, flow_type
        )

    async def resend_code_post(
        self,
        device_id: str,
        pre_auth_session_id: str,
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        ResendCodePostOkResult, ResendCodePostRestartFlowError, GeneralErrorResponse
    ]:
        device_info = await api_options.recipe_implementation.list_codes_by_device_id(
            device_id=device_id, user_context=user_context
        )
        if device_info is None:
            return ResendCodePostRestartFlowError()
        if (
            isinstance(api_options.config.contact_config, ContactEmailOnlyConfig)
            and device_info.email is None
        ) or (
            isinstance(api_options.config.contact_config, ContactPhoneOnlyConfig)
            and device_info.phone_number is None
        ):
            return ResendCodePostRestartFlowError()
        number_of_tries_to_create_new_code = 0
        while True:
            number_of_tries_to_create_new_code += 1
            user_input_code = None
            if api_options.config.get_custom_user_input_code is not None:
                user_input_code = await api_options.config.get_custom_user_input_code(
                    user_context
                )
            response = (
                await api_options.recipe_implementation.create_new_code_for_device(
                    device_id=device_id,
                    user_input_code=user_input_code,
                    user_context=user_context,
                )
            )
            if isinstance(
                response, CreateNewCodeForDeviceUserInputCodeAlreadyUsedError
            ):
                if number_of_tries_to_create_new_code >= 3:
                    return GeneralErrorResponse(
                        "Failed to generate a one time code. Please try again"
                    )
                continue

            if isinstance(response, CreateNewCodeForDeviceOkResult):
                magic_link = None
                user_input_code = None
                flow_type = api_options.config.flow_type
                if flow_type in ("MAGIC_LINK", "USER_INPUT_CODE_AND_MAGIC_LINK"):
                    magic_link = (
                        api_options.app_info.website_domain.get_as_string_dangerous()
                        + api_options.app_info.website_base_path.get_as_string_dangerous()
                        + "/verify"
                        + "?rid="
                        + api_options.recipe_id
                        + "&preAuthSessionId="
                        + response.pre_auth_session_id
                        + "#"
                        + response.link_code
                    )
                if flow_type in ("USER_INPUT_CODE", "USER_INPUT_CODE_AND_MAGIC_LINK"):
                    user_input_code = response.user_input_code

                if isinstance(
                    api_options.config.contact_config, ContactEmailOnlyConfig
                ) or (
                    isinstance(
                        api_options.config.contact_config, ContactEmailOrPhoneConfig
                    )
                    and device_info.email is not None
                ):
                    if device_info.email is None:
                        raise Exception("Should never come here")

                    log_debug_message(
                        "Sending passwordless login email to %s", device_info.email
                    )
                    passwordless_email_delivery_input = (
                        PasswordlessLoginEmailTemplateVars(
                            email=device_info.email,
                            user_input_code=user_input_code,
                            url_with_link_code=magic_link,
                            code_life_time=response.code_life_time,
                            pre_auth_session_id=response.pre_auth_session_id,
                        )
                    )
                    await api_options.email_delivery.ingredient_interface_impl.send_email(
                        passwordless_email_delivery_input, user_context
                    )
                elif isinstance(
                    api_options.config.contact_config,
                    (ContactEmailOrPhoneConfig, ContactPhoneOnlyConfig),
                ):
                    if device_info.phone_number is None:
                        raise Exception("Should never come here")
                    log_debug_message(
                        "Sending passwordless login SMS to %s", device_info.phone_number
                    )
                    sms_input = PasswordlessLoginSMSTemplateVars(
                        phone_number=device_info.phone_number,
                        user_input_code=user_input_code,
                        url_with_link_code=magic_link,
                        code_life_time=response.code_life_time,
                        pre_auth_session_id=response.pre_auth_session_id,
                    )
                    await api_options.sms_delivery.ingredient_interface_impl.send_sms(
                        sms_input, user_context
                    )
                return ResendCodePostOkResult()
            return ResendCodePostRestartFlowError()

    async def consume_code_post(
        self,
        pre_auth_session_id: str,
        user_input_code: Union[str, None],
        device_id: Union[str, None],
        link_code: Union[str, None],
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        ConsumeCodePostOkResult,
        ConsumeCodePostRestartFlowError,
        ConsumeCodePostIncorrectUserInputCodeError,
        ConsumeCodePostExpiredUserInputCodeError,
        GeneralErrorResponse,
    ]:
        response = await api_options.recipe_implementation.consume_code(
            pre_auth_session_id=pre_auth_session_id,
            user_input_code=user_input_code,
            device_id=device_id,
            link_code=link_code,
            user_context=user_context,
        )

        if isinstance(response, ConsumeCodeExpiredUserInputCodeError):
            return ConsumeCodePostExpiredUserInputCodeError(
                failed_code_input_attempt_count=response.failed_code_input_attempt_count,
                maximum_code_input_attempts=response.maximum_code_input_attempts,
            )
        if isinstance(response, ConsumeCodeIncorrectUserInputCodeError):
            return ConsumeCodePostIncorrectUserInputCodeError(
                failed_code_input_attempt_count=response.failed_code_input_attempt_count,
                maximum_code_input_attempts=response.maximum_code_input_attempts,
            )
        if isinstance(response, ConsumeCodeRestartFlowError):
            return ConsumeCodePostRestartFlowError()

        user = response.user

        if user.email is not None:
            ev_instance = EmailVerificationRecipe.get_instance_optional()
            if ev_instance is not None:
                token_response = await ev_instance.recipe_implementation.create_email_verification_token(
                    user.user_id,
                    user.email,
                    user_context,
                )

                if isinstance(token_response, CreateEmailVerificationTokenOkResult):
                    await ev_instance.recipe_implementation.verify_email_using_token(
                        token_response.token,
                        user_context,
                    )

        session = await create_new_session(
            api_options.request,
            user.user_id,
            {},
            {},
            user_context=user_context,
        )

        return ConsumeCodePostOkResult(
            created_new_user=response.created_new_user,
            user=response.user,
            session=session,
        )

    async def email_exists_get(
        self, email: str, api_options: APIOptions, user_context: Dict[str, Any]
    ) -> Union[EmailExistsGetOkResult, GeneralErrorResponse]:
        response = await api_options.recipe_implementation.get_user_by_email(
            email, user_context
        )
        return EmailExistsGetOkResult(exists=response is not None)

    async def phone_number_exists_get(
        self, phone_number: str, api_options: APIOptions, user_context: Dict[str, Any]
    ) -> Union[PhoneNumberExistsGetOkResult, GeneralErrorResponse]:
        response = await api_options.recipe_implementation.get_user_by_phone_number(
            phone_number, user_context
        )
        return PhoneNumberExistsGetOkResult(exists=response is not None)



"""
Copyright (c) 2020, VRAI Labs and/or its affiliates. All rights reserved.

This software is licensed under the Apache License, Version 2.0 (the
"License") as published by the Apache Software Foundation.

You may not use this file except in compliance with the License. You may
obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations
under the License.
"""

from typing import Any, Dict, List, Union

from _pytest.fixtures import fixture
from fastapi import FastAPI
from pytest import mark
from starlette.requests import Request
from starlette.testclient import TestClient
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import jwt
from supertokens_python.recipe.jwt.asyncio import create_jwt
from supertokens_python.recipe.jwt.interfaces import (
    APIInterface,
    APIOptions,
    CreateJwtOkResult,
    RecipeInterface,
)
from tests.utils import clean_st, reset, setup_st, start_st


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    @app.post("/jwtcreate")
    async def jwt_create(request: Request):  # type: ignore
        payload = (await request.json())["payload"]
        response = await create_jwt(payload, 1000)
        return response

    return TestClient(app)


@mark.asyncio
async def test_that_default_getJWKS_api_does_not_work_when_disabled(
    driver_config_client: TestClient,
):
    created_jwt = None
    jwt_keys: List[Dict[str, Any]] = []

    def custom_functions(param: RecipeInterface):
        temp = param.get_jwks

        async def get_jwks(user_context: Dict[str, Any]):
            response_ = await temp(user_context)

            for key in response_.keys:
                jwt_keys.append(
                    {
                        "kty": key.kty,
                        "kid": key.kid,
                        "n": key.n,
                        "e": key.e,
                        "alg": key.alg,
                        "use": key.use,
                    }
                )

            return response_

        temp1 = param.create_jwt

        async def create_jwt_(
            payload: Dict[str, Any],
            validity_seconds: Union[int, None],
            user_context: Dict[str, Any],
        ):
            response_ = await temp1(payload, validity_seconds, user_context)

            if isinstance(response_, CreateJwtOkResult):
                nonlocal created_jwt
                created_jwt = response_.jwt

            return response_

        param.create_jwt = create_jwt_
        param.get_jwks = get_jwks
        return param

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[jwt.init(override=jwt.OverrideConfig(functions=custom_functions))],
    )
    start_st()

    response = driver_config_client.post(
        url="/jwtcreate", json={"payload": {"someKey": "key"}}
    )

    assert response is not None
    assert response.json()["jwt"] == created_jwt

    response = driver_config_client.get(url="/auth/jwt/jwks.json")

    assert response is not None
    assert response.json()["keys"] == jwt_keys


@mark.asyncio
async def test_overriding_APIs(driver_config_client: TestClient):
    jwt_keys = None

    def custom_api(param: APIInterface):
        temp = param.jwks_get

        async def get_jwks_get(api_options: APIOptions, user_context: Dict[str, Any]):
            response_ = await temp(api_options, user_context)
            nonlocal jwt_keys
            jwt_keys = response_.to_json()["keys"]
            return response_

        param.jwks_get = get_jwks_get

        return param

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[jwt.init(override=jwt.OverrideConfig(apis=custom_api))],
    )
    start_st()

    response = driver_config_client.get(url="/auth/jwt/jwks.json")

    assert response is not None
    assert response.json()["keys"] == jwt_keys



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, List, Union

from supertokens_python.ingredients.emaildelivery import EmailDeliveryIngredient
from supertokens_python.recipe.emailpassword.types import EmailTemplateVars
from ...supertokens import AppInfo

from ...types import APIResponse, GeneralErrorResponse

if TYPE_CHECKING:
    from supertokens_python.framework import BaseRequest, BaseResponse
    from supertokens_python.recipe.session import SessionContainer

    from .types import FormField, User
    from .utils import EmailPasswordConfig


class SignUpOkResult:
    def __init__(self, user: User):
        self.user = user


class SignUpEmailAlreadyExistsError:
    pass


class SignInOkResult:
    def __init__(self, user: User):
        self.user = user


class SignInWrongCredentialsError:
    pass


class CreateResetPasswordOkResult:
    def __init__(self, token: str):
        self.token = token


class CreateResetPasswordWrongUserIdError:
    pass


class ResetPasswordUsingTokenOkResult:
    def __init__(self, user_id: Union[str, None]):
        self.user_id = user_id


class ResetPasswordUsingTokenInvalidTokenError:
    pass


class UpdateEmailOrPasswordOkResult:
    pass


class UpdateEmailOrPasswordEmailAlreadyExistsError:
    pass


class UpdateEmailOrPasswordUnknownUserIdError:
    pass


class RecipeInterface(ABC):
    def __init__(self):
        pass

    @abstractmethod
    async def get_user_by_id(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        pass

    @abstractmethod
    async def get_user_by_email(
        self, email: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        pass

    @abstractmethod
    async def create_reset_password_token(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[CreateResetPasswordOkResult, CreateResetPasswordWrongUserIdError]:
        pass

    @abstractmethod
    async def reset_password_using_token(
        self, token: str, new_password: str, user_context: Dict[str, Any]
    ) -> Union[
        ResetPasswordUsingTokenOkResult, ResetPasswordUsingTokenInvalidTokenError
    ]:
        pass

    @abstractmethod
    async def sign_in(
        self, email: str, password: str, user_context: Dict[str, Any]
    ) -> Union[SignInOkResult, SignInWrongCredentialsError]:
        pass

    @abstractmethod
    async def sign_up(
        self, email: str, password: str, user_context: Dict[str, Any]
    ) -> Union[SignUpOkResult, SignUpEmailAlreadyExistsError]:
        pass

    @abstractmethod
    async def update_email_or_password(
        self,
        user_id: str,
        email: Union[str, None],
        password: Union[str, None],
        user_context: Dict[str, Any],
    ) -> Union[
        UpdateEmailOrPasswordOkResult,
        UpdateEmailOrPasswordEmailAlreadyExistsError,
        UpdateEmailOrPasswordUnknownUserIdError,
    ]:
        pass


class APIOptions:
    def __init__(
        self,
        request: BaseRequest,
        response: BaseResponse,
        recipe_id: str,
        config: EmailPasswordConfig,
        recipe_implementation: RecipeInterface,
        app_info: AppInfo,
        email_delivery: EmailDeliveryIngredient[EmailTemplateVars],
    ):
        self.request: BaseRequest = request
        self.response: BaseResponse = response
        self.recipe_id: str = recipe_id
        self.config: EmailPasswordConfig = config
        self.recipe_implementation: RecipeInterface = recipe_implementation
        self.app_info = app_info
        self.email_delivery = email_delivery


class EmailExistsGetOkResult(APIResponse):
    status: str = "OK"

    def __init__(self, exists: bool):
        self.exists = exists

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status, "exists": self.exists}


class GeneratePasswordResetTokenPostOkResult(APIResponse):
    status: str = "OK"

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status}


class PasswordResetPostOkResult(APIResponse):
    status: str = "OK"

    def __init__(self, user_id: Union[str, None]):
        self.user_id = user_id

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status}


class PasswordResetPostInvalidTokenResponse(APIResponse):
    status: str = "RESET_PASSWORD_INVALID_TOKEN_ERROR"

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status}


class SignInPostOkResult(APIResponse):
    status: str = "OK"

    def __init__(self, user: User, session: SessionContainer):
        self.user = user
        self.session = session

    def to_json(self) -> Dict[str, Any]:
        return {
            "status": self.status,
            "user": {
                "id": self.user.user_id,
                "email": self.user.email,
                "timeJoined": self.user.time_joined,
            },
        }


class SignInPostWrongCredentialsError(APIResponse):
    status: str = "WRONG_CREDENTIALS_ERROR"

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status}


class SignUpPostOkResult(APIResponse):
    status: str = "OK"

    def __init__(self, user: User, session: SessionContainer):
        self.user = user
        self.session = session

    def to_json(self) -> Dict[str, Any]:
        return {
            "status": self.status,
            "user": {
                "id": self.user.user_id,
                "email": self.user.email,
                "timeJoined": self.user.time_joined,
            },
        }


class SignUpPostEmailAlreadyExistsError(APIResponse):
    status: str = "EMAIL_ALREADY_EXISTS_ERROR"

    def to_json(self) -> Dict[str, Any]:
        return {"status": self.status}


class APIInterface:
    def __init__(self):
        self.disable_email_exists_get = False
        self.disable_generate_password_reset_token_post = False
        self.disable_password_reset_post = False
        self.disable_sign_in_post = False
        self.disable_sign_up_post = False

    @abstractmethod
    async def email_exists_get(
        self, email: str, api_options: APIOptions, user_context: Dict[str, Any]
    ) -> Union[EmailExistsGetOkResult, GeneralErrorResponse]:
        pass

    @abstractmethod
    async def generate_password_reset_token_post(
        self,
        form_fields: List[FormField],
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[GeneratePasswordResetTokenPostOkResult, GeneralErrorResponse]:
        pass

    @abstractmethod
    async def password_reset_post(
        self,
        form_fields: List[FormField],
        token: str,
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        PasswordResetPostOkResult,
        PasswordResetPostInvalidTokenResponse,
        GeneralErrorResponse,
    ]:
        pass

    @abstractmethod
    async def sign_in_post(
        self,
        form_fields: List[FormField],
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        SignInPostOkResult, SignInPostWrongCredentialsError, GeneralErrorResponse
    ]:
        pass

    @abstractmethod
    async def sign_up_post(
        self,
        form_fields: List[FormField],
        api_options: APIOptions,
        user_context: Dict[str, Any],
    ) -> Union[
        SignUpPostOkResult, SignUpPostEmailAlreadyExistsError, GeneralErrorResponse
    ]:
        pass



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Union

if TYPE_CHECKING:
    from .interfaces import APIInterface, RecipeInterface


class OverrideConfig:
    def __init__(
        self,
        functions: Union[Callable[[RecipeInterface], RecipeInterface], None] = None,
        apis: Union[Callable[[APIInterface], APIInterface], None] = None,
    ):
        self.functions = functions
        self.apis = apis


class JWTConfig:
    def __init__(self, override: OverrideConfig, jwt_validity_seconds: int):
        self.override = override
        self.jwt_validity_seconds = jwt_validity_seconds


def validate_and_normalise_user_input(
    jwt_validity_seconds: Union[int, None] = None,
    override: Union[OverrideConfig, None] = None,
):
    if jwt_validity_seconds is not None and not isinstance(jwt_validity_seconds, int):  # type: ignore
        raise ValueError("jwt_validity_seconds must be an integer or None")

    if override is not None and not isinstance(override, OverrideConfig):  # type: ignore
        raise ValueError("override must be an instance of OverrideConfig or None")

    if override is None:
        override = OverrideConfig()
    if jwt_validity_seconds is None:
        jwt_validity_seconds = 3153600000

    return JWTConfig(override, jwt_validity_seconds)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Dict, List, Union

from pytest import fixture, mark
from supertokens_python import InputAppInfo, SupertokensConfig, init
from supertokens_python.framework.fastapi import get_middleware
from supertokens_python.recipe import emailpassword, session
from supertokens_python.recipe.emailpassword.asyncio import sign_up
from supertokens_python.recipe.emailpassword.interfaces import (
    APIInterface,
    APIOptions,
    RecipeInterface,
)
from supertokens_python.recipe.emailpassword.types import FormField
from supertokens_python.recipe.session.interfaces import (
    RecipeInterface as SRecipeInterface,
)

from fastapi import FastAPI
from fastapi.testclient import TestClient

from .utils import clean_st, reset, setup_st, sign_in_request, start_st

works = False
signUpContextWorks = False


def setup_function(_):
    reset()
    clean_st()
    setup_st()


def teardown_function(_):
    reset()
    clean_st()


@fixture(scope="function")
async def driver_config_client():
    app = FastAPI()
    app.add_middleware(get_middleware())

    return TestClient(app)


@mark.asyncio
async def test_user_context(driver_config_client: TestClient):
    global works
    global signUpContextWorks

    def apis_override_email_password(param: APIInterface):
        og_sign_in_post = param.sign_in_post

        async def sign_in_post(
            form_fields: List[FormField],
            api_options: APIOptions,
            user_context: Dict[str, Any],
        ):
            user_context = {"preSignInPOST": True}
            response = await og_sign_in_post(form_fields, api_options, user_context)
            if (
                "preSignInPOST" in user_context
                and "preSignIn" in user_context
                and "preCreateNewSession" in user_context
                and "postCreateNewSession" in user_context
                and "postSignIn" in user_context
            ):
                global works
                works = True
            return response

        param.sign_in_post = sign_in_post
        return param

    def functions_override_email_password(param: RecipeInterface):
        og_sign_in = param.sign_in
        og_sign_up = param.sign_up

        async def sign_up_(email: str, password: str, user_context: Dict[str, Any]):
            if "manualCall" in user_context:
                global signUpContextWorks
                signUpContextWorks = True
            response = await og_sign_up(email, password, user_context)
            return response

        async def sign_in(email: str, password: str, user_context: Dict[str, Any]):
            if "preSignInPOST" in user_context:
                user_context["preSignIn"] = True
            response = await og_sign_in(email, password, user_context)
            if "preSignInPOST" in user_context and "preSignIn" in user_context:
                user_context["postSignIn"] = True
            return response

        param.sign_in = sign_in
        param.sign_up = sign_up_
        return param

    def functions_override_session(param: SRecipeInterface):
        og_create_new_session = param.create_new_session

        async def create_new_session(
            request: Any,
            user_id: str,
            access_token_payload: Union[None, Dict[str, Any]],
            session_data: Union[None, Dict[str, Any]],
            user_context: Dict[str, Any],
        ):
            if (
                "preSignInPOST" in user_context
                and "preSignIn" in user_context
                and "postSignIn" in user_context
            ):
                user_context["preCreateNewSession"] = True
            response = await og_create_new_session(
                request, user_id, access_token_payload, session_data, user_context
            )
            if (
                "preSignInPOST" in user_context
                and "preSignIn" in user_context
                and "postSignIn" in user_context
                and "preCreateNewSession" in user_context
            ):
                user_context["postCreateNewSession"] = True
            return response

        param.create_new_session = create_new_session
        return param

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(
                override=emailpassword.InputOverrideConfig(
                    apis=apis_override_email_password,
                    functions=functions_override_email_password,
                )
            ),
            session.init(
                override=session.InputOverrideConfig(
                    functions=functions_override_session
                )
            ),
        ],
    )
    start_st()

    await sign_up("random@gmail.com", "validpass123", {"manualCall": True})

    res = sign_in_request(driver_config_client, "random@gmail.com", "validpass123")
    assert res.status_code == 200
    assert works
    assert signUpContextWorks


@mark.asyncio
async def test_default_context(driver_config_client: TestClient):
    signin_api_context_works, signin_context_works, create_new_session_context_works = (
        False,
        False,
        False,
    )

    def apis_override_email_password(param: APIInterface):
        og_sign_in_post = param.sign_in_post

        async def sign_in_post(
            form_fields: List[FormField],
            api_options: APIOptions,
            user_context: Dict[str, Any],
        ):
            req = user_context.get("_default", {}).get("request")
            if req:
                nonlocal signin_api_context_works
                signin_api_context_works = True

            return await og_sign_in_post(form_fields, api_options, user_context)

        param.sign_in_post = sign_in_post
        return param

    def functions_override_email_password(param: RecipeInterface):
        og_sign_in = param.sign_in

        async def sign_in(email: str, password: str, user_context: Dict[str, Any]):
            req = user_context.get("_default", {}).get("request")
            if req:
                nonlocal signin_context_works
                signin_context_works = True

            return await og_sign_in(email, password, user_context)

        param.sign_in = sign_in
        return param

    def functions_override_session(param: SRecipeInterface):
        og_create_new_session = param.create_new_session

        async def create_new_session(
            request: Any,
            user_id: str,
            access_token_payload: Union[None, Dict[str, Any]],
            session_data: Union[None, Dict[str, Any]],
            user_context: Dict[str, Any],
        ):
            req = user_context.get("_default", {}).get("request")
            if req:
                nonlocal create_new_session_context_works
                create_new_session_context_works = True

            response = await og_create_new_session(
                request, user_id, access_token_payload, session_data, user_context
            )
            return response

        param.create_new_session = create_new_session
        return param

    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain="http://api.supertokens.io",
            website_domain="http://supertokens.io",
        ),
        framework="fastapi",
        recipe_list=[
            emailpassword.init(
                override=emailpassword.InputOverrideConfig(
                    apis=apis_override_email_password,
                    functions=functions_override_email_password,
                )
            ),
            session.init(
                override=session.InputOverrideConfig(
                    functions=functions_override_session
                )
            ),
        ],
    )
    start_st()

    await sign_up("random@gmail.com", "validpass123", {"manualCall": True})
    res = sign_in_request(driver_config_client, "random@gmail.com", "validpass123")

    assert res.status_code == 200
    assert all(
        [
            signin_api_context_works,
            signin_context_works,
            create_new_session_context_works,
        ]
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from typing import Any, Dict, List, Union

from supertokens_python.normalised_url_path import NormalisedURLPath
from supertokens_python.querier import Querier

from .interfaces import (
    AddRoleToUserOkResult,
    CreateNewRoleOrAddPermissionsOkResult,
    DeleteRoleOkResult,
    GetAllRolesOkResult,
    GetPermissionsForRoleOkResult,
    GetRolesForUserOkResult,
    GetRolesThatHavePermissionOkResult,
    GetUsersThatHaveRoleOkResult,
    RecipeInterface,
    RemovePermissionsFromRoleOkResult,
    RemoveUserRoleOkResult,
    UnknownRoleError,
)


class RecipeImplementation(RecipeInterface):
    def __init__(self, querier: Querier):
        super().__init__()
        self.querier = querier

    async def add_role_to_user(
        self, user_id: str, role: str, user_context: Dict[str, Any]
    ) -> Union[AddRoleToUserOkResult, UnknownRoleError]:
        params = {"userId": user_id, "role": role}
        response = await self.querier.send_put_request(
            NormalisedURLPath("/recipe/user/role"), params
        )
        if response.get("status") == "OK":
            return AddRoleToUserOkResult(
                did_user_already_have_role=response["didUserAlreadyHaveRole"]
            )
        return UnknownRoleError()

    async def remove_user_role(
        self, user_id: str, role: str, user_context: Dict[str, Any]
    ) -> Union[RemoveUserRoleOkResult, UnknownRoleError]:
        params = {"userId": user_id, "role": role}
        response = await self.querier.send_post_request(
            NormalisedURLPath("/recipe/user/role/remove"), params
        )
        if response["status"] == "OK":
            return RemoveUserRoleOkResult(
                did_user_have_role=response["didUserHaveRole"]
            )
        return UnknownRoleError()

    async def get_roles_for_user(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> GetRolesForUserOkResult:
        params = {"userId": user_id}
        response = await self.querier.send_get_request(
            NormalisedURLPath("/recipe/user/roles"), params
        )
        return GetRolesForUserOkResult(roles=response["roles"])

    async def get_users_that_have_role(
        self, role: str, user_context: Dict[str, Any]
    ) -> Union[GetUsersThatHaveRoleOkResult, UnknownRoleError]:
        params = {"role": role}
        response = await self.querier.send_get_request(
            NormalisedURLPath("/recipe/role/users"), params
        )
        if response.get("status") == "OK":
            return GetUsersThatHaveRoleOkResult(users=response["users"])
        return UnknownRoleError()

    async def create_new_role_or_add_permissions(
        self, role: str, permissions: List[str], user_context: Dict[str, Any]
    ) -> CreateNewRoleOrAddPermissionsOkResult:
        params = {"role": role, "permissions": permissions}
        response = await self.querier.send_put_request(
            NormalisedURLPath("/recipe/role"), params
        )
        return CreateNewRoleOrAddPermissionsOkResult(
            created_new_role=response["createdNewRole"]
        )

    async def get_permissions_for_role(
        self, role: str, user_context: Dict[str, Any]
    ) -> Union[GetPermissionsForRoleOkResult, UnknownRoleError]:
        params = {"role": role}
        response = await self.querier.send_get_request(
            NormalisedURLPath("/recipe/role/permissions"), params
        )
        if response.get("status") == "OK":
            return GetPermissionsForRoleOkResult(permissions=response["permissions"])
        return UnknownRoleError()

    async def remove_permissions_from_role(
        self, role: str, permissions: List[str], user_context: Dict[str, Any]
    ) -> Union[RemovePermissionsFromRoleOkResult, UnknownRoleError]:
        params = {"role": role, "permissions": permissions}
        response = await self.querier.send_post_request(
            NormalisedURLPath("/recipe/role/permissions/remove"), params
        )
        if response.get("status") == "OK":
            return RemovePermissionsFromRoleOkResult()
        return UnknownRoleError()

    async def get_roles_that_have_permission(
        self, permission: str, user_context: Dict[str, Any]
    ) -> GetRolesThatHavePermissionOkResult:
        params = {"permission": permission}
        response = await self.querier.send_get_request(
            NormalisedURLPath("/recipe/permission/roles"), params
        )
        return GetRolesThatHavePermissionOkResult(roles=response["roles"])

    async def delete_role(
        self, role: str, user_context: Dict[str, Any]
    ) -> DeleteRoleOkResult:
        params = {"role": role}
        response = await self.querier.send_post_request(
            NormalisedURLPath("/recipe/role/remove"), params
        )
        return DeleteRoleOkResult(did_role_exist=response["didRoleExist"])

    async def get_all_roles(self, user_context: Dict[str, Any]) -> GetAllRolesOkResult:
        params = {}
        response = await self.querier.send_get_request(
            NormalisedURLPath("/recipe/roles"), params
        )
        return GetAllRolesOkResult(roles=response["roles"])



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, Awaitable, List

from supertokens_python.supertokens import Supertokens

from ...usermetadata import UserMetadataRecipe
from ...usermetadata.asyncio import get_user_metadata
from ..interfaces import DashboardUsersGetResponse
from ..utils import UserWithMetadata

if TYPE_CHECKING:
    from supertokens_python.recipe.dashboard.interfaces import (
        APIOptions,
        APIInterface,
    )
    from supertokens_python.types import APIResponse

from supertokens_python.exceptions import GeneralError, raise_bad_input_exception


async def handle_users_get_api(
    api_implementation: APIInterface, api_options: APIOptions
) -> APIResponse:
    _ = api_implementation

    limit = api_options.request.get_query_param("limit")
    if limit is None:
        raise_bad_input_exception("Missing required parameter 'limit'")

    time_joined_order: str = api_options.request.get_query_param(  # type: ignore
        "timeJoinedOrder", "DESC"
    )
    if time_joined_order not in ["ASC", "DESC"]:
        raise_bad_input_exception("Invalid value recieved for 'timeJoinedOrder'")

    pagination_token = api_options.request.get_query_param("paginationToken")

    users_response = await Supertokens.get_instance().get_users(
        limit=int(limit),
        time_joined_order=time_joined_order,  # type: ignore
        pagination_token=pagination_token,
        include_recipe_ids=None,
    )

    # user metadata bulk fetch with batches:

    try:
        UserMetadataRecipe.get_instance()
    except GeneralError:
        return DashboardUsersGetResponse(
            users_response.users, users_response.next_pagination_token
        )

    users_with_metadata: List[UserWithMetadata] = [
        UserWithMetadata().from_user(user) for user in users_response.users
    ]
    metadata_fetch_awaitables: List[Awaitable[Any]] = []

    async def get_user_metadata_and_update_user(user_idx: int) -> None:
        user = users_response.users[user_idx]
        user_metadata = await get_user_metadata(user.user_id)
        first_name = user_metadata.metadata.get("first_name")
        last_name = user_metadata.metadata.get("last_name")

        # None becomes null which is acceptable for the dashboard.
        users_with_metadata[user_idx].first_name = first_name
        users_with_metadata[user_idx].last_name = last_name

    # Batch calls to get user metadata:
    for i, _ in enumerate(users_response.users):
        metadata_fetch_awaitables.append(get_user_metadata_and_update_user(i))

    promise_arr_start_position = 0
    batch_size = 5

    while promise_arr_start_position < len(metadata_fetch_awaitables):
        # We want to query only 5 in parallel at a time
        promises_to_call = [
            metadata_fetch_awaitables[i]
            for i in range(
                promise_arr_start_position,
                min(
                    promise_arr_start_position + batch_size,
                    len(metadata_fetch_awaitables),
                ),
            )
        ]
        await asyncio.gather(*promises_to_call)

        promise_arr_start_position += batch_size

    return DashboardUsersGetResponse(
        users_with_metadata,
        users_response.next_pagination_token,
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Dict, List, Union

from supertokens_python.recipe.passwordless.interfaces import (
    ConsumeCodeExpiredUserInputCodeError,
    ConsumeCodeIncorrectUserInputCodeError,
    ConsumeCodeOkResult,
    ConsumeCodeRestartFlowError,
    CreateCodeOkResult,
    CreateNewCodeForDeviceOkResult,
    CreateNewCodeForDeviceRestartFlowError,
    CreateNewCodeForDeviceUserInputCodeAlreadyUsedError,
    DeleteUserInfoOkResult,
    DeleteUserInfoUnknownUserIdError,
    RevokeAllCodesOkResult,
    RevokeCodeOkResult,
    UpdateUserEmailAlreadyExistsError,
    UpdateUserOkResult,
    UpdateUserPhoneNumberAlreadyExistsError,
    UpdateUserUnknownUserIdError,
)
from supertokens_python.recipe.passwordless.recipe import PasswordlessRecipe
from supertokens_python.recipe.passwordless.types import (
    DeviceType,
    EmailTemplateVars,
    SMSTemplateVars,
    User,
)


async def create_code(
    email: Union[None, str] = None,
    phone_number: Union[None, str] = None,
    user_input_code: Union[None, str] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> CreateCodeOkResult:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.create_code(
        email=email,
        phone_number=phone_number,
        user_input_code=user_input_code,
        user_context=user_context,
    )


async def create_new_code_for_device(
    device_id: str,
    user_input_code: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[
    CreateNewCodeForDeviceOkResult,
    CreateNewCodeForDeviceRestartFlowError,
    CreateNewCodeForDeviceUserInputCodeAlreadyUsedError,
]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.create_new_code_for_device(
        device_id=device_id, user_input_code=user_input_code, user_context=user_context
    )


async def consume_code(
    pre_auth_session_id: str,
    user_input_code: Union[str, None] = None,
    device_id: Union[str, None] = None,
    link_code: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[
    ConsumeCodeOkResult,
    ConsumeCodeIncorrectUserInputCodeError,
    ConsumeCodeExpiredUserInputCodeError,
    ConsumeCodeRestartFlowError,
]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.consume_code(
        pre_auth_session_id=pre_auth_session_id,
        user_input_code=user_input_code,
        device_id=device_id,
        link_code=link_code,
        user_context=user_context,
    )


async def get_user_by_id(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[User, None]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.get_user_by_id(
        user_id=user_id, user_context=user_context
    )


async def get_user_by_email(
    email: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[User, None]:
    if user_context is None:
        user_context = {}
    return (
        await PasswordlessRecipe.get_instance().recipe_implementation.get_user_by_email(
            email=email, user_context=user_context
        )
    )


async def get_user_by_phone_number(
    phone_number: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[User, None]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.get_user_by_phone_number(
        phone_number=phone_number, user_context=user_context
    )


async def update_user(
    user_id: str,
    email: Union[str, None] = None,
    phone_number: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[
    UpdateUserOkResult,
    UpdateUserUnknownUserIdError,
    UpdateUserEmailAlreadyExistsError,
    UpdateUserPhoneNumberAlreadyExistsError,
]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.update_user(
        user_id=user_id,
        email=email,
        phone_number=phone_number,
        user_context=user_context,
    )


async def delete_email_for_user(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[DeleteUserInfoOkResult, DeleteUserInfoUnknownUserIdError]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.delete_email_for_user(
        user_id=user_id, user_context=user_context
    )


async def delete_phone_number_for_user(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[DeleteUserInfoOkResult, DeleteUserInfoUnknownUserIdError]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.delete_phone_number_for_user(
        user_id=user_id, user_context=user_context
    )


async def revoke_all_codes(
    email: Union[str, None] = None,
    phone_number: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> RevokeAllCodesOkResult:
    if user_context is None:
        user_context = {}
    return (
        await PasswordlessRecipe.get_instance().recipe_implementation.revoke_all_codes(
            email=email, phone_number=phone_number, user_context=user_context
        )
    )


async def revoke_code(
    code_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> RevokeCodeOkResult:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.revoke_code(
        code_id=code_id, user_context=user_context
    )


async def list_codes_by_email(
    email: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[DeviceType]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.list_codes_by_email(
        email=email, user_context=user_context
    )


async def list_codes_by_phone_number(
    phone_number: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[DeviceType]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.list_codes_by_phone_number(
        phone_number=phone_number, user_context=user_context
    )


async def list_codes_by_device_id(
    device_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[DeviceType, None]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.list_codes_by_device_id(
        device_id=device_id, user_context=user_context
    )


async def list_codes_by_pre_auth_session_id(
    pre_auth_session_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[DeviceType, None]:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().recipe_implementation.list_codes_by_pre_auth_session_id(
        pre_auth_session_id=pre_auth_session_id, user_context=user_context
    )


async def create_magic_link(
    email: Union[str, None],
    phone_number: Union[str, None],
    user_context: Union[None, Dict[str, Any]] = None,
) -> str:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().create_magic_link(
        email=email, phone_number=phone_number, user_context=user_context
    )


async def signinup(
    email: Union[str, None],
    phone_number: Union[str, None],
    user_context: Union[None, Dict[str, Any]] = None,
) -> ConsumeCodeOkResult:
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().signinup(
        email=email, phone_number=phone_number, user_context=user_context
    )


async def send_email(
    input_: EmailTemplateVars, user_context: Union[None, Dict[str, Any]] = None
):
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().email_delivery.ingredient_interface_impl.send_email(
        input_, user_context
    )


async def send_sms(
    input_: SMSTemplateVars, user_context: Union[None, Dict[str, Any]] = None
):
    if user_context is None:
        user_context = {}
    return await PasswordlessRecipe.get_instance().sms_delivery.ingredient_interface_impl.send_sms(
        input_, user_context
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import List, Union

from supertokens_python.ingredients.emaildelivery import EmailDeliveryIngredient
from supertokens_python.ingredients.emaildelivery.types import (
    SMTPServiceInterface,
    EmailDeliveryInterface,
)
from supertokens_python.ingredients.smsdelivery import SMSDeliveryIngredient
from supertokens_python.ingredients.smsdelivery.types import (
    TwilioServiceInterface,
    SMSDeliveryInterface,
)


class User:
    def __init__(
        self,
        user_id: str,
        email: Union[str, None],
        phone_number: Union[str, None],
        time_joined: int,
    ):
        self.user_id: str = user_id
        self.email: Union[str, None] = email
        self.phone_number: Union[str, None] = phone_number
        self.time_joined: int = time_joined


class DeviceCode:
    def __init__(self, code_id: str, time_created: str, code_life_time: int):
        self.code_id = code_id
        self.time_created = time_created
        self.code_life_time = code_life_time


class DeviceType:
    def __init__(
        self,
        pre_auth_session_id: str,
        failed_code_input_attempt_count: int,
        codes: List[DeviceCode],
        email: Union[str, None] = None,
        phone_number: Union[str, None] = None,
    ):
        self.pre_auth_session_id = pre_auth_session_id
        self.failed_code_input_attempt_count = failed_code_input_attempt_count
        self.codes = codes
        self.email = email
        self.phone_number = phone_number


class CreateAndSendCustomEmailParameters:
    def __init__(
        self,
        code_life_time: int,
        pre_auth_session_id: str,
        email: str,
        user_input_code: Union[str, None] = None,
        url_with_link_code: Union[str, None] = None,
    ):
        self.email: str = email
        self.code_life_time: int = code_life_time
        self.pre_auth_session_id: str = pre_auth_session_id
        self.user_input_code: Union[str, None] = user_input_code
        self.url_with_link_code: Union[str, None] = url_with_link_code


PasswordlessLoginEmailTemplateVars = CreateAndSendCustomEmailParameters


class CreateAndSendCustomTextMessageParameters:
    def __init__(
        self,
        code_life_time: int,
        pre_auth_session_id: str,
        phone_number: str,
        user_input_code: Union[str, None] = None,
        url_with_link_code: Union[str, None] = None,
    ):
        self.code_life_time: int = code_life_time
        self.pre_auth_session_id: str = pre_auth_session_id
        self.phone_number: str = phone_number
        self.user_input_code: Union[str, None] = user_input_code
        self.url_with_link_code: Union[str, None] = url_with_link_code


PasswordlessLoginSMSTemplateVars = CreateAndSendCustomTextMessageParameters


# Export:
EmailTemplateVars = PasswordlessLoginEmailTemplateVars
SMSTemplateVars = PasswordlessLoginSMSTemplateVars

SMTPOverrideInput = SMTPServiceInterface[EmailTemplateVars]
TwilioOverrideInput = TwilioServiceInterface[SMSTemplateVars]

EmailDeliveryOverrideInput = EmailDeliveryInterface[EmailTemplateVars]
SMSDeliveryOverrideInput = SMSDeliveryInterface[SMSTemplateVars]


class PasswordlessIngredients:
    def __init__(
        self,
        email_delivery: Union[EmailDeliveryIngredient[EmailTemplateVars], None] = None,
        sms_delivery: Union[SMSDeliveryIngredient[SMSTemplateVars], None] = None,
    ):
        self.email_delivery = email_delivery
        self.sms_delivery = sms_delivery



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, List, Union

from ...passwordless.interfaces import (
    ConsumeCodeExpiredUserInputCodeError,
    ConsumeCodeIncorrectUserInputCodeError,
)
from ...passwordless.interfaces import (
    ConsumeCodeOkResult as PasswordlessConsumeCodeOkResult,
)
from ...passwordless.interfaces import (
    ConsumeCodeRestartFlowError,
    CreateCodeOkResult,
    CreateNewCodeForDeviceOkResult,
    CreateNewCodeForDeviceRestartFlowError,
    CreateNewCodeForDeviceUserInputCodeAlreadyUsedError,
    DeleteUserInfoOkResult,
    DeleteUserInfoUnknownUserIdError,
    DeviceType,
    RevokeAllCodesOkResult,
    RevokeCodeOkResult,
    UpdateUserEmailAlreadyExistsError,
    UpdateUserOkResult,
    UpdateUserPhoneNumberAlreadyExistsError,
    UpdateUserUnknownUserIdError,
)
from ...thirdparty.interfaces import SignInUpOkResult

if TYPE_CHECKING:
    from supertokens_python.querier import Querier

from supertokens_python.recipe.passwordless.recipe_implementation import (
    RecipeImplementation as PasswordlessImplementation,
)
from supertokens_python.recipe.thirdparty.recipe_implementation import (
    RecipeImplementation as ThirdPartyImplementation,
)

from ..interfaces import ConsumeCodeOkResult, RecipeInterface
from ..types import User
from .passwordless_recipe_implementation import (
    RecipeImplementation as DerivedPasswordlessImplementation,
)
from .third_party_recipe_implementation import (
    RecipeImplementation as DerivedThirdPartyImplementation,
)


class RecipeImplementation(RecipeInterface):
    def __init__(
        self, passwordless_querier: Querier, thirdparty_querier: Union[Querier, None]
    ):
        super().__init__()
        passwordless_implementation = PasswordlessImplementation(passwordless_querier)

        self.pless_get_user_by_id = passwordless_implementation.get_user_by_id
        self.pless_get_user_by_email = passwordless_implementation.get_user_by_email
        self.pless_consume_code = passwordless_implementation.consume_code
        self.pless_create_code = passwordless_implementation.create_code
        self.pless_create_new_code_for_device = (
            passwordless_implementation.create_new_code_for_device
        )
        self.pless_get_user_by_phone_number = (
            passwordless_implementation.get_user_by_phone_number
        )
        self.pless_list_codes_by_device_id = (
            passwordless_implementation.list_codes_by_device_id
        )
        self.pless_list_codes_by_email = passwordless_implementation.list_codes_by_email
        self.pless_list_codes_by_phone_number = (
            passwordless_implementation.list_codes_by_phone_number
        )
        self.pless_list_codes_by_pre_auth_session_id = (
            passwordless_implementation.list_codes_by_pre_auth_session_id
        )
        self.pless_revoke_all_codes = passwordless_implementation.revoke_all_codes
        self.pless_revoke_code = passwordless_implementation.revoke_code
        self.pless_update_user = passwordless_implementation.update_user
        self.pless_delete_email_for_user = (
            passwordless_implementation.delete_email_for_user
        )
        self.pless_delete_phone_number_for_user = (
            passwordless_implementation.delete_phone_number_for_user
        )

        derived_pless = DerivedPasswordlessImplementation(self)
        passwordless_implementation.get_user_by_id = derived_pless.get_user_by_id
        passwordless_implementation.get_user_by_email = derived_pless.get_user_by_email
        passwordless_implementation.consume_code = derived_pless.consume_code
        passwordless_implementation.create_code = derived_pless.create_code
        passwordless_implementation.create_new_code_for_device = (
            derived_pless.create_new_code_for_device
        )
        passwordless_implementation.get_user_by_phone_number = (
            derived_pless.get_user_by_phone_number
        )
        passwordless_implementation.list_codes_by_device_id = (
            derived_pless.list_codes_by_device_id
        )
        passwordless_implementation.list_codes_by_email = (
            derived_pless.list_codes_by_email
        )
        passwordless_implementation.list_codes_by_phone_number = (
            derived_pless.list_codes_by_phone_number
        )
        passwordless_implementation.list_codes_by_pre_auth_session_id = (
            derived_pless.list_codes_by_pre_auth_session_id
        )
        passwordless_implementation.revoke_all_codes = derived_pless.revoke_all_codes
        passwordless_implementation.revoke_code = derived_pless.revoke_code
        passwordless_implementation.update_user = derived_pless.update_user
        passwordless_implementation.delete_email_for_user = (
            derived_pless.delete_email_for_user
        )
        passwordless_implementation.delete_phone_number_for_user = (
            derived_pless.delete_phone_number_for_user
        )

        self.tp_get_user_by_id = None
        self.tp_get_users_by_email = None
        self.tp_get_user_by_thirdparty_info = None
        self.tp_sign_in_up = None
        if thirdparty_querier is not None:
            thirdparty_implementation = ThirdPartyImplementation(thirdparty_querier)
            self.tp_get_user_by_id = thirdparty_implementation.get_user_by_id
            self.tp_get_users_by_email = thirdparty_implementation.get_users_by_email
            self.tp_get_user_by_thirdparty_info = (
                thirdparty_implementation.get_user_by_thirdparty_info
            )
            self.tp_sign_in_up = thirdparty_implementation.sign_in_up

            derived_tp = DerivedThirdPartyImplementation(self)
            thirdparty_implementation.get_user_by_id = derived_tp.get_user_by_id
            thirdparty_implementation.get_users_by_email = derived_tp.get_users_by_email
            thirdparty_implementation.get_user_by_thirdparty_info = (
                derived_tp.get_user_by_thirdparty_info
            )
            thirdparty_implementation.sign_in_up = derived_tp.sign_in_up

    async def get_user_by_id(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        pless_user = await self.pless_get_user_by_id(user_id, user_context)

        if pless_user is not None:
            return User(
                user_id=pless_user.user_id,
                email=pless_user.email,
                time_joined=pless_user.time_joined,
                third_party_info=None,
                phone_number=pless_user.phone_number,
            )

        if self.tp_get_user_by_id is None:
            return None

        tp_user = await self.tp_get_user_by_id(user_id, user_context)
        if tp_user is None:
            return None
        return User(
            user_id=tp_user.user_id,
            email=tp_user.email,
            time_joined=tp_user.time_joined,
            third_party_info=tp_user.third_party_info,
            phone_number=None,
        )

    async def get_users_by_email(
        self, email: str, user_context: Dict[str, Any]
    ) -> List[User]:
        result: List[User] = []
        pless_user = await self.pless_get_user_by_email(email, user_context)

        if pless_user is not None:
            result.append(
                User(
                    user_id=pless_user.user_id,
                    email=pless_user.email,
                    time_joined=pless_user.time_joined,
                    third_party_info=None,
                    phone_number=pless_user.phone_number,
                )
            )

        if self.tp_get_users_by_email is None:
            return result

        tp_users = await self.tp_get_users_by_email(email, user_context)

        for tp_user in tp_users:
            result.append(
                User(
                    user_id=tp_user.user_id,
                    email=tp_user.email,
                    time_joined=tp_user.time_joined,
                    third_party_info=tp_user.third_party_info,
                    phone_number=None,
                )
            )

        return result

    async def get_user_by_thirdparty_info(
        self,
        third_party_id: str,
        third_party_user_id: str,
        user_context: Dict[str, Any],
    ) -> Union[User, None]:
        if self.tp_get_user_by_thirdparty_info is None:
            return None
        tp_user = await self.tp_get_user_by_thirdparty_info(
            third_party_id, third_party_user_id, user_context
        )

        if tp_user is None:
            return None

        return User(
            user_id=tp_user.user_id,
            email=tp_user.email,
            time_joined=tp_user.time_joined,
            third_party_info=tp_user.third_party_info,
            phone_number=None,
        )

    async def thirdparty_sign_in_up(
        self,
        third_party_id: str,
        third_party_user_id: str,
        email: str,
        user_context: Dict[str, Any],
    ) -> SignInUpOkResult:
        if self.tp_sign_in_up is None:
            raise Exception("No thirdparty provider configured")
        return await self.tp_sign_in_up(
            third_party_id, third_party_user_id, email, user_context
        )

    async def get_user_by_phone_number(
        self, phone_number: str, user_context: Dict[str, Any]
    ) -> Union[User, None]:
        pless_user = await self.pless_get_user_by_phone_number(
            phone_number, user_context
        )
        if pless_user is not None:
            return User(
                user_id=pless_user.user_id,
                email=pless_user.email,
                time_joined=pless_user.time_joined,
                third_party_info=None,
                phone_number=pless_user.phone_number,
            )

        return None

    async def create_code(
        self,
        email: Union[None, str],
        phone_number: Union[None, str],
        user_input_code: Union[None, str],
        user_context: Dict[str, Any],
    ) -> CreateCodeOkResult:
        return await self.pless_create_code(
            email, phone_number, user_input_code, user_context
        )

    async def create_new_code_for_device(
        self,
        device_id: str,
        user_input_code: Union[str, None],
        user_context: Dict[str, Any],
    ) -> Union[
        CreateNewCodeForDeviceOkResult,
        CreateNewCodeForDeviceRestartFlowError,
        CreateNewCodeForDeviceUserInputCodeAlreadyUsedError,
    ]:
        return await self.pless_create_new_code_for_device(
            device_id, user_input_code, user_context
        )

    async def consume_code(
        self,
        pre_auth_session_id: str,
        user_input_code: Union[str, None],
        device_id: Union[str, None],
        link_code: Union[str, None],
        user_context: Dict[str, Any],
    ) -> Union[
        ConsumeCodeOkResult,
        ConsumeCodeIncorrectUserInputCodeError,
        ConsumeCodeExpiredUserInputCodeError,
        ConsumeCodeRestartFlowError,
    ]:
        result = await self.pless_consume_code(
            pre_auth_session_id, user_input_code, device_id, link_code, user_context
        )
        if isinstance(result, PasswordlessConsumeCodeOkResult):
            return ConsumeCodeOkResult(
                result.created_new_user,
                User(
                    result.user.user_id,
                    result.user.email,
                    result.user.phone_number,
                    None,
                    result.user.time_joined,
                ),
            )
        return result

    async def update_passwordless_user(
        self,
        user_id: str,
        email: Union[str, None],
        phone_number: Union[str, None],
        user_context: Dict[str, Any],
    ) -> Union[
        UpdateUserOkResult,
        UpdateUserUnknownUserIdError,
        UpdateUserEmailAlreadyExistsError,
        UpdateUserPhoneNumberAlreadyExistsError,
    ]:
        user = await self.get_user_by_id(user_id, user_context)
        if user is None:
            return UpdateUserUnknownUserIdError()
        if user.third_party_info is not None:
            raise Exception(
                "Cannot update passwordless user info of a user who signed up using third party login."
            )
        return await self.pless_update_user(user_id, email, phone_number, user_context)

    async def delete_email_for_passwordless_user(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[DeleteUserInfoOkResult, DeleteUserInfoUnknownUserIdError]:
        user = await self.get_user_by_id(user_id, user_context)
        if user is None:
            return DeleteUserInfoUnknownUserIdError()
        if user.third_party_info is not None:
            raise Exception(
                "Cannot update passwordless user info of a user who signed up using third party login."
            )
        return await self.pless_delete_email_for_user(user_id, user_context)

    async def delete_phone_number_for_user(
        self, user_id: str, user_context: Dict[str, Any]
    ) -> Union[DeleteUserInfoOkResult, DeleteUserInfoUnknownUserIdError]:
        user = await self.get_user_by_id(user_id, user_context)
        if user is None:
            return DeleteUserInfoUnknownUserIdError()
        if user.third_party_info is not None:
            raise Exception(
                "Cannot update passwordless user info of a user who signed up using third party login."
            )
        return await self.pless_delete_phone_number_for_user(user_id, user_context)

    async def revoke_all_codes(
        self,
        email: Union[str, None],
        phone_number: Union[str, None],
        user_context: Dict[str, Any],
    ) -> RevokeAllCodesOkResult:
        return await self.pless_revoke_all_codes(email, phone_number, user_context)

    async def revoke_code(
        self, code_id: str, user_context: Dict[str, Any]
    ) -> RevokeCodeOkResult:
        return await self.pless_revoke_code(code_id, user_context)

    async def list_codes_by_email(
        self, email: str, user_context: Dict[str, Any]
    ) -> List[DeviceType]:
        return await self.pless_list_codes_by_email(email, user_context)

    async def list_codes_by_phone_number(
        self, phone_number: str, user_context: Dict[str, Any]
    ) -> List[DeviceType]:
        return await self.pless_list_codes_by_phone_number(phone_number, user_context)

    async def list_codes_by_device_id(
        self, device_id: str, user_context: Dict[str, Any]
    ) -> Union[DeviceType, None]:
        return await self.pless_list_codes_by_device_id(device_id, user_context)

    async def list_codes_by_pre_auth_session_id(
        self, pre_auth_session_id: str, user_context: Dict[str, Any]
    ) -> Union[DeviceType, None]:
        return await self.pless_list_codes_by_pre_auth_session_id(
            pre_auth_session_id, user_context
        )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from abc import ABC, abstractmethod
from typing import Any, Dict, Union

from supertokens_python.framework import BaseRequest, BaseResponse
from supertokens_python.recipe.jwt.interfaces import (
    CreateJwtOkResult,
    CreateJwtResultUnsupportedAlgorithm,
    GetJWKSResult,
)
from supertokens_python.types import APIResponse, GeneralErrorResponse

from .utils import OpenIdConfig


class GetOpenIdDiscoveryConfigurationResult:
    def __init__(self, issuer: str, jwks_uri: str):
        self.issuer = issuer
        self.jwks_uri = jwks_uri


class RecipeInterface(ABC):
    def __init__(self):
        pass

    @abstractmethod
    async def create_jwt(
        self,
        payload: Dict[str, Any],
        validity_seconds: Union[int, None],
        user_context: Dict[str, Any],
    ) -> Union[CreateJwtOkResult, CreateJwtResultUnsupportedAlgorithm]:
        pass

    @abstractmethod
    async def get_jwks(self, user_context: Dict[str, Any]) -> GetJWKSResult:
        pass

    @abstractmethod
    async def get_open_id_discovery_configuration(
        self, user_context: Dict[str, Any]
    ) -> GetOpenIdDiscoveryConfigurationResult:
        pass


class APIOptions:
    def __init__(
        self,
        request: BaseRequest,
        response: BaseResponse,
        recipe_id: str,
        config: OpenIdConfig,
        recipe_implementation: RecipeInterface,
    ):
        self.request = request
        self.response = response
        self.recipe_id = recipe_id
        self.config = config
        self.recipe_implementation = recipe_implementation


class OpenIdDiscoveryConfigurationGetResponse(APIResponse):
    status: str = "OK"

    def __init__(self, issuer: str, jwks_uri: str):
        self.issuer = issuer
        self.jwks_uri = jwks_uri

    def to_json(self):
        return {"status": self.status, "issuer": self.issuer, "jwks_uri": self.jwks_uri}


class APIInterface:
    def __init__(self):
        self.disable_open_id_discovery_configuration_get = False

    @abstractmethod
    async def open_id_discovery_configuration_get(
        self, api_options: APIOptions, user_context: Dict[str, Any]
    ) -> Union[OpenIdDiscoveryConfigurationGetResponse, GeneralErrorResponse]:
        pass



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from os import environ
from typing import List
from enum import Enum


class AllowedProcessStates(Enum):
    CALLING_SERVICE_IN_VERIFY = 1
    CALLING_SERVICE_IN_GET_HANDSHAKE_INFO = 2
    CALLING_SERVICE_IN_GET_API_VERSION = 3
    CALLING_SERVICE_IN_REQUEST_HELPER = 4


class ProcessState:
    __instance = None

    def __init__(self):
        self.history: List[AllowedProcessStates] = []

    @staticmethod
    def get_instance():
        if ProcessState.__instance is None:
            ProcessState.__instance = ProcessState()
        return ProcessState.__instance

    def add_state(self, state: AllowedProcessStates):
        if ("SUPERTOKENS_ENV" in environ) and (environ["SUPERTOKENS_ENV"] == "testing"):
            self.history.append(state)

    def reset(self):
        self.history = []



import json
from typing import Any, Dict

from supertokens_python.exceptions import raise_bad_input_exception
from supertokens_python.recipe.usermetadata import UserMetadataRecipe
from supertokens_python.recipe.usermetadata.asyncio import (
    clear_user_metadata,
    update_user_metadata,
)

from ...interfaces import APIInterface, APIOptions, UserMetadataPutAPIResponse


async def handle_metadata_put(
    _api_interface: APIInterface, api_options: APIOptions
) -> UserMetadataPutAPIResponse:
    request_body: Dict[str, Any] = await api_options.request.json()  # type: ignore
    user_id = request_body.get("userId")
    data = request_body.get("data")

    # This is to throw an error early in case the recipe has not been initialised
    UserMetadataRecipe.get_instance()

    if user_id is None or isinstance(user_id, str) is False:
        raise_bad_input_exception(
            "Required parameter 'userId' is missing or has an invalid type"
        )

    if data is None or isinstance(data, str) is False:
        raise_bad_input_exception(
            "Required parameter 'data' is missing or has an invalid type"
        )

    parsed_data: Dict[str, Any] = {}
    try:
        parsed_data = json.loads(data)
        if not isinstance(parsed_data, dict):  # type: ignore
            raise Exception()

    except Exception:
        raise_bad_input_exception("'data' must be a valid JSON body")

    # This API is meant to set the user metadata of a user. We delete the existing data
    # before updating it because we want to make sure that shallow merging does not result
    # in the data being incorrect
    #
    # For example if the old data is {test: "test", test2: "test2"} and the user wants to delete
    # test2 from the data simply calling updateUserMetadata with {test: "test"} would not remove
    # test2 because of shallow merging.
    #
    # Removing first ensures that the final data is exactly what the user wanted it to be

    await clear_user_metadata(user_id)
    await update_user_metadata(user_id, parsed_data)

    return UserMetadataPutAPIResponse()



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, Callable, Union

from supertokens_python.ingredients.emaildelivery.services.smtp import Transporter
from supertokens_python.ingredients.emaildelivery.types import (
    EmailDeliveryInterface,
    SMTPServiceInterface,
    SMTPSettings,
)
from supertokens_python.recipe.emailpassword.types import (
    EmailTemplateVars,
    SMTPOverrideInput,
)

from .service_implementation import ServiceImplementation


class SMTPService(EmailDeliveryInterface[EmailTemplateVars]):
    service_implementation: SMTPServiceInterface[EmailTemplateVars]

    def __init__(
        self,
        smtp_settings: SMTPSettings,
        override: Union[Callable[[SMTPOverrideInput], SMTPOverrideInput], None] = None,
    ) -> None:
        transporter = Transporter(smtp_settings)

        oi = ServiceImplementation(transporter)
        self.service_implementation = oi if override is None else override(oi)

    async def send_email(
        self, template_vars: EmailTemplateVars, user_context: Dict[str, Any]
    ) -> None:
        content = await self.service_implementation.get_content(
            template_vars, user_context
        )
        await self.service_implementation.send_raw_email(content, user_context)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing_extensions import Literal
from typing import Callable, List, Union

from . import supertokens
from .recipe_module import RecipeModule

InputAppInfo = supertokens.InputAppInfo
Supertokens = supertokens.Supertokens
SupertokensConfig = supertokens.SupertokensConfig
AppInfo = supertokens.AppInfo


def init(
    app_info: InputAppInfo,
    framework: Literal["fastapi", "flask", "django"],
    supertokens_config: SupertokensConfig,
    recipe_list: List[Callable[[supertokens.AppInfo], RecipeModule]],
    mode: Union[Literal["asgi", "wsgi"], None] = None,
    telemetry: Union[bool, None] = None,
):
    return Supertokens.init(
        app_info, framework, supertokens_config, recipe_list, mode, telemetry
    )


def get_all_cors_headers() -> List[str]:
    return supertokens.Supertokens.get_instance().get_all_cors_headers()



from pytest import mark

from supertokens_python import InputAppInfo, Supertokens, SupertokensConfig, init
from supertokens_python.recipe import session
from supertokens_python.recipe.session import SessionRecipe


@mark.parametrize(
    "api_domain,website_domain,cookie_same_site",
    [
        ("https://foo.com/", "https://bar.com/", "none"),  # Different URLs
        (
            "https://foo.example.com/",
            "https://bar.example.com/",
            "lax",
        ),  # Different subdomains
        (
            "https://foo.azurewebsites.net/",
            "https://bar.azurewebsites.net/",
            "none",
        ),  # PSL so same_site should be none despite
        (
            "http://foo.example.com/",
            "http://foo.example.com/",
            "lax",
        ),  # HTTP same url and subdomain
        ("http://example.com/", "http://example.com/", "lax"),  # HTTP same url
    ],
)
def test_same_site_cookie_values(
    api_domain: str, website_domain: str, cookie_same_site: str
):
    init(
        supertokens_config=SupertokensConfig("http://localhost:3567"),
        app_info=InputAppInfo(
            app_name="SuperTokens Demo",
            api_domain=api_domain,
            website_domain=website_domain,
            api_base_path="/auth",
        ),
        framework="fastapi",
        recipe_list=[
            session.init(
                cookie_same_site="strict" if cookie_same_site == "strict" else None
            ),
        ],
    )

    s = SessionRecipe.get_instance()
    assert s.config.cookie_same_site == cookie_same_site
    s.reset()
    Supertokens.reset()



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import Union, Dict, Any

from supertokens_python.ingredients.emaildelivery import EmailDeliveryIngredient
from supertokens_python.ingredients.emaildelivery.types import (
    SMTPServiceInterface,
    EmailDeliveryInterface,
)


class User:
    def __init__(self, user_id: str, email: str):
        self.user_id = user_id
        self.email = email


class VerificationEmailTemplateVarsUser:
    def __init__(self, user_id: str, email: str):
        self.id = user_id
        self.email = email


class VerificationEmailTemplateVars:
    def __init__(
        self,
        user: VerificationEmailTemplateVarsUser,
        email_verify_link: str,
        user_context: Dict[str, Any],
    ) -> None:
        self.user = user
        self.email_verify_link = email_verify_link
        self.user_context = user_context


# Export:
EmailTemplateVars = VerificationEmailTemplateVars

SMTPOverrideInput = SMTPServiceInterface[EmailTemplateVars]

EmailDeliveryOverrideInput = EmailDeliveryInterface[EmailTemplateVars]


class EmailVerificationIngredients:
    def __init__(
        self,
        email_delivery: Union[EmailDeliveryIngredient[EmailTemplateVars], None] = None,
    ):
        self.email_delivery = email_delivery



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import abc
from typing import TYPE_CHECKING, List, Union

from typing_extensions import Literal

from .framework.response import BaseResponse

if TYPE_CHECKING:
    from supertokens_python.framework.request import BaseRequest
    from .supertokens import AppInfo
    from .normalised_url_path import NormalisedURLPath

from .exceptions import SuperTokensError


class RecipeModule(abc.ABC):
    def __init__(self, recipe_id: str, app_info: AppInfo):
        self.recipe_id = recipe_id
        self.app_info = app_info

    def get_recipe_id(self):
        return self.recipe_id

    def get_app_info(self):
        return self.app_info

    def return_api_id_if_can_handle_request(
        self, path: NormalisedURLPath, method: str
    ) -> Union[str, None]:
        apis_handled = self.get_apis_handled()
        for current_api in apis_handled:
            if (
                not current_api.disabled
                and current_api.method == method
                and self.app_info.api_base_path.append(
                    current_api.path_without_api_base_path
                ).equals(path)
            ):
                return current_api.request_id
        return None

    @abc.abstractmethod
    def is_error_from_this_recipe_based_on_instance(self, err: Exception) -> bool:
        pass

    @abc.abstractmethod
    def get_apis_handled(self) -> List[APIHandled]:
        pass

    @abc.abstractmethod
    async def handle_api_request(
        self,
        request_id: str,
        request: BaseRequest,
        path: NormalisedURLPath,
        method: str,
        response: BaseResponse,
    ) -> Union[BaseResponse, None]:
        pass

    @abc.abstractmethod
    async def handle_error(
        self, request: BaseRequest, err: SuperTokensError, response: BaseResponse
    ) -> BaseResponse:
        pass

    @abc.abstractmethod
    def get_all_cors_headers(self) -> List[str]:
        pass


class APIHandled:
    def __init__(
        self,
        path_without_api_base_path: NormalisedURLPath,
        method: Literal["post", "get", "delete", "put", "options", "trace"],
        request_id: str,
        disabled: bool,
    ):
        self.path_without_api_base_path = path_without_api_base_path
        self.method = method
        self.request_id = request_id
        self.disabled = disabled



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from supertokens_python.recipe.dashboard.interfaces import (
        APIOptions,
        APIInterface,
    )

from supertokens_python.utils import (
    send_200_response,
    send_non_200_response_with_message,
)

from ..utils import validate_api_key


async def handle_validate_key_api(
    _api_implementation: APIInterface, api_options: APIOptions
):

    is_valid_key = validate_api_key(api_options.request, api_options.config)

    if is_valid_key:
        return send_200_response({"status": "OK"}, api_options.response)
    return send_non_200_response_with_message("Unauthorised", 401, api_options.response)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Dict
from supertokens_python.recipe.session.syncio import update_access_token_payload
from supertokens_python.recipe.session.interfaces import APIInterface, RecipeInterface
from supertokens_python.framework import BaseRequest, BaseResponse
import json
import os
import sys
from functools import wraps
from typing import Union

from typing_extensions import Literal

from flask import Flask, g, jsonify, make_response, render_template, request
from flask.wrappers import Response
from flask_cors import CORS
from supertokens_python import InputAppInfo, Supertokens, SupertokensConfig, init
from supertokens_python.framework.flask.flask_middleware import Middleware
from supertokens_python.recipe import session
from supertokens_python.recipe.session import (
    InputErrorHandlers,
    SessionContainer,
    SessionRecipe,
)
from supertokens_python.recipe.session.framework.flask import verify_session
from supertokens_python.recipe.session.syncio import (
    create_new_session,
    revoke_all_sessions_for_user,
)

last_set_enable_anti_csrf = True
last_set_enable_jwt = False

index_file = open("templates/index.html", "r")
file_contents = index_file.read()
index_file.close()

app = Flask(__name__, template_folder="templates")
Middleware(app)
CORS(app, supports_credentials=True)
os.environ.setdefault("SUPERTOKENS_ENV", "testing")


def custom_decorator_for_test():  # type: ignore
    def session_verify_custom_test(f):  # type: ignore
        @wraps(f)  # type: ignore
        def wrapped_function(*args, **kwargs):  # type: ignore
            Test.increment_attempted_refresh()
            try:
                value: Response = f(*args, **kwargs)
                if value is not None and value.status_code != 200:
                    return value
                if request.headers.get("rid") is None:  # type: ignore
                    return "refresh failed"
                Test.increment_refresh()
                return "refresh success"
            except Exception as e:
                raise e

        return wrapped_function  # type: ignore

    return session_verify_custom_test  # type: ignore


def try_refresh_token(_):
    return jsonify({"error": "try refresh token"}), 401


def unauthorised(_):
    return jsonify({"error": "unauthorised"}), 401


class Test:
    no_of_times_refresh_called_during_test = 0
    no_of_times_get_session_called_during_test = 0
    no_of_times_refresh_attempted_during_test = 0

    @staticmethod
    def reset():
        Test.no_of_times_refresh_called_during_test = 0
        Test.no_of_times_get_session_called_during_test = 0
        Test.no_of_times_refresh_attempted_during_test = 0

    @staticmethod
    def increment_refresh():
        Test.no_of_times_refresh_called_during_test = (
            Test.no_of_times_refresh_called_during_test + 1
        )

    @staticmethod
    def increment_attempted_refresh():
        Test.no_of_times_refresh_attempted_during_test = (
            Test.no_of_times_refresh_attempted_during_test + 1
        )

    @staticmethod
    def increment_get_session():
        Test.no_of_times_get_session_called_during_test = (
            Test.no_of_times_get_session_called_during_test + 1
        )

    @staticmethod
    def get_session_called_count():
        return Test.no_of_times_get_session_called_during_test

    @staticmethod
    def get_refresh_called_count():
        return Test.no_of_times_refresh_called_during_test

    @staticmethod
    def get_refresh_attempted_count():
        return Test.no_of_times_refresh_attempted_during_test


async def unauthorised_f(_: BaseRequest, __: str, res: BaseResponse):
    res.set_status_code(401)
    res.set_json_content({})
    return res


def apis_override_session(param: APIInterface):
    param.disable_refresh_post = True
    return param


def functions_override_session(param: RecipeInterface):
    original_create_new_session = param.create_new_session

    async def create_new_session_custom(
        _request: BaseRequest,
        user_id: str,
        access_token_payload: Union[Dict[str, Any], None],
        session_data: Union[Dict[str, Any], None],
        user_context: Dict[str, Any],
    ) -> SessionContainer:
        if access_token_payload is None:
            access_token_payload = {}
        access_token_payload = {**access_token_payload, "customClaim": "customValue"}
        return await original_create_new_session(
            _request, user_id, access_token_payload, session_data, user_context
        )

    param.create_new_session = create_new_session_custom

    return param


def get_app_port():
    argvv = sys.argv
    for i in range(0, len(argvv)):  # pylint: disable=consider-using-enumerate
        if argvv[i] == "--port":
            return argvv[i + 1]

    return "8080"


def config(
    enable_anti_csrf: bool, enable_jwt: bool, jwt_property_name: Union[str, None]
):
    anti_csrf: Literal["VIA_TOKEN", "NONE"] = "NONE"
    if enable_anti_csrf:
        anti_csrf = "VIA_TOKEN"
    if enable_jwt:
        init(
            supertokens_config=SupertokensConfig("http://localhost:9000"),
            app_info=InputAppInfo(
                app_name="SuperTokens Python SDK",
                api_domain="0.0.0.0:" + get_app_port(),
                website_domain="http://localhost.org:8080",
            ),
            framework="flask",
            recipe_list=[
                session.init(
                    error_handlers=InputErrorHandlers(on_unauthorised=unauthorised_f),
                    anti_csrf=anti_csrf,
                    override=session.InputOverrideConfig(
                        apis=apis_override_session, functions=functions_override_session
                    ),
                    jwt=session.JWTConfig(enable_jwt, jwt_property_name),
                )
            ],
            telemetry=False,
        )
    else:
        init(
            supertokens_config=SupertokensConfig("http://localhost:9000"),
            app_info=InputAppInfo(
                app_name="SuperTokens Python SDK",
                api_domain="0.0.0.0:" + get_app_port(),
                website_domain="http://localhost.org:8080",
            ),
            framework="flask",
            recipe_list=[
                session.init(
                    error_handlers=InputErrorHandlers(on_unauthorised=unauthorised_f),
                    anti_csrf=anti_csrf,
                    override=session.InputOverrideConfig(apis=apis_override_session),
                )
            ],
            telemetry=False,
        )


config(True, False, None)


@app.route("/index.html", methods=["GET"])  # type: ignore
def send_file():
    return render_template("index.html")


from flask import send_from_directory  # type: ignore


@app.route("/angular/<path:path>", methods=["GET"])  # type: ignore
def send_angular_file(path: str):
    return send_from_directory("templates/angular", path)


def send_options_api_response():
    return ""


@app.route("/login", methods=["OPTIONS"])  # type: ignore
def login_options():
    return send_options_api_response()


@app.route("/login", methods=["POST"])  # type: ignore
def login():
    user_id: str = request.get_json()["userId"]  # type: ignore
    _session = create_new_session(request, user_id)
    return _session.get_user_id()


@app.route("/beforeeach", methods=["OPTIONS"])  # type: ignore
def before_each_options():
    return send_options_api_response()


@app.route("/beforeeach", methods=["POST"])  # type: ignore
def before_each():
    Test.reset()
    return ""


@app.route("/testUserConfig", methods=["OPTIONS"])  # type: ignore
def test_user_config_options():
    return send_options_api_response()


@app.route("/testUserConfig", methods=["POST"])  # type: ignore
def test_config():
    return ""


@app.route("/multipleInterceptors", methods=["OPTIONS"])  # type: ignore
def multiple_interceptors_options():
    return send_options_api_response()


@app.route("/multipleInterceptors", methods=["POST"])  # type: ignore
def multiple_interceptors():
    result_bool = "success" if "interceptorheader2" in request.headers and "interceptorheader1" in request.headers else "failure"  # type: ignore
    return str(result_bool)


@app.route("/", methods=["OPTIONS"])  # type: ignore
def options():
    return send_options_api_response()


@app.route("/", methods=["GET"])  # type: ignore
@verify_session()
def get_info():
    Test.increment_get_session()
    session = g.supertokens
    resp = make_response(session.get_user_id())
    resp.headers["Cache-Control"] = "no-cache, private"
    return resp


@app.route("/update-jwt", methods=["OPTIONS"])  # type: ignore
def update_options():
    return send_options_api_response()


@app.route("/update-jwt", methods=["GET"])  # type: ignore
@verify_session()
# @supertokens_middleware(True)
def update_jwt():
    Test.increment_get_session()
    _session = g.supertokens

    resp = make_response(_session.get_access_token_payload())
    resp.headers["Cache-Control"] = "no-cache, private"
    return resp


@app.route("/update-jwt", methods=["POST"])  # type: ignore
@verify_session()
# @supertokens_middleware()
def update_jwt_post():
    _session = g.supertokens
    _session.sync_update_access_token_payload(request.get_json())
    Test.increment_get_session()
    resp = make_response(_session.get_access_token_payload())
    resp.headers["Cache-Control"] = "no-cache, private"
    return resp


@app.route("/update-jwt-with-handle", methods=["POST"])  # type: ignore
@verify_session()
def update_jwt_with_handle_post():
    _session = g.supertokens
    update_access_token_payload(_session.get_handle(), request.get_json())
    resp = make_response(_session.get_access_token_payload())
    resp.headers["Cache-Control"] = "no-cache, private"
    return resp


@app.route("/testing", methods=["OPTIONS"])  # type: ignore
def testing_options():
    return send_options_api_response()


@app.route("/testing", methods=["GET"])  # type: ignore
def testing():
    if "testing" in request.headers:  # type: ignore
        resp = make_response("success")
        resp.headers["testing"] = request.headers["testing"]  # type: ignore
        return resp
    return "success"


@app.route("/testing", methods=["PUT"])  # type: ignore
def testing_put():
    if "testing" in request.headers:  # type: ignore
        resp = make_response("success")
        resp.headers["testing"] = request.headers["testing"]  # type: ignore
        return resp
    return "success"


@app.route("/testing", methods=["POST"])  # type: ignore
def testing_post():
    if "testing" in request.headers:  # type: ignore
        resp = make_response("success")
        resp.headers["testing"] = request.headers["testing"]  # type: ignore
        return resp
    return "success"


@app.route("/testing", methods=["DELETE"])  # type: ignore
def testing_delete():
    if "testing" in request.headers:  # type: ignore
        resp = make_response("success")
        resp.headers["testing"] = request.headers["testing"]  # type: ignore
        return resp
    return "success"


@app.route("/logout", methods=["OPTIONS"])  # type: ignore
def logout_options():
    return send_options_api_response()


@app.route("/logout", methods=["POST"])  # type: ignore
@verify_session()
def logout():
    _session = g.supertokens

    _session.sync_revoke_session()
    return "success"


@app.route("/revokeAll", methods=["OPTIONS"])  # type: ignore
def revoke_all_options():
    return send_options_api_response()


@app.route("/revokeAll", methods=["POST"])  # type: ignore
@verify_session()
async def revoke_all():
    session_ = g.supertokens
    revoke_all_sessions_for_user(session_.get_user_id())
    return "success"


@app.route("/refresh", methods=["OPTIONS"])  # type: ignore
def refresh_options():
    return send_options_api_response()


@app.route("/refreshAttemptedTime", methods=["GET"])  # type: ignore
def refresh_attempted_time():
    return str(Test.get_refresh_attempted_count())


@app.route("/auth/session/refresh", methods=["POST"])  # type: ignore
@custom_decorator_for_test()
@verify_session()
def refresh():
    return ""


@app.route("/setAntiCsrf", methods=["POST"])  # type: ignore
def set_anti_csrf():
    global last_set_enable_anti_csrf
    json: Dict[str, Any] = request.get_json(silent=True)  # type: ignore
    if "enableAntiCsrf" not in json:
        enable_csrf = True
    else:
        enable_csrf = json["enableAntiCsrf"]

    last_set_enable_anti_csrf = enable_csrf
    if enable_csrf is not None:
        Supertokens.reset()
        SessionRecipe.reset()
        config(enable_csrf, False, None)
    return "success", 200


@app.route("/setEnableJWT", methods=["POST"])  # type: ignore
def set_enable_jwt():
    global last_set_enable_jwt
    global last_set_enable_anti_csrf  # pylint: disable=global-variable-not-assigned
    json: Dict[str, Any] = request.get_json(silent=True)  # type: ignore
    if "enableJWT" not in json:
        enable_jwt = False
    else:
        enable_jwt = json["enableJWT"]

    last_set_enable_jwt = enable_jwt
    if enable_jwt is not None:
        Supertokens.reset()
        SessionRecipe.reset()
        config(last_set_enable_anti_csrf, enable_jwt, None)
    return "success", 200


@app.route("/refreshCalledTime", methods=["OPTIONS"])  # type: ignore
def refresh_called_time_options():
    return send_options_api_response()


@app.route("/refreshCalledTime", methods=["GET"])  # type: ignore
def refresh_called_time():
    return str(Test.get_refresh_called_count())


@app.route("/getSessionCalledTime", methods=["OPTIONS"])  # type: ignore
def get_session_called_time_options():
    return send_options_api_response()


@app.route("/getSessionCalledTime", methods=["GET"])  # type: ignore
def get_session_called_time():
    return str(Test.get_session_called_count())


@app.route("/ping", methods=["OPTIONS"])  # type: ignore
def ping_options():
    return send_options_api_response()


@app.route("/ping", methods=["GET"])  # type: ignore
def ping():
    return "success"


@app.route("/testHeader", methods=["OPTIONS"])  # type: ignore
def test_header_options():
    return send_options_api_response()


@app.route("/testHeader", methods=["GET"])  # type: ignore
def test_header():  # type: ignore
    success_info = request.headers.get("st-custom-header")  # type: ignore
    return {"success": success_info}  # type: ignore


@app.route("/checkDeviceInfo", methods=["OPTIONS"])  # type: ignore
def check_device_info_options():
    return send_options_api_response()


@app.route("/checkDeviceInfo", methods=["GET"])  # type: ignore
def check_device_info():
    sdk_name = request.headers.get("supertokens-sdk-name")  # type: ignore
    sdk_version = request.headers.get("supertokens-sdk-version")  # type: ignore
    return "true" if sdk_name == "website" and isinstance(sdk_version, str) else "false"


@app.route("/check-rid", methods=["GET"])  # type: ignore
def check_rid():
    rid = request.headers.get("rid")  # type: ignore

    return "fail" if rid is None else "success"


@app.route("/featureFlags", methods=["GET"])  # type: ignore
def feature_flags():
    global last_set_enable_jwt  # pylint: disable=global-variable-not-assigned

    return jsonify({"sessionJwt": last_set_enable_jwt})


@app.route("/reinitialiseBackendConfig", methods=["POST"])  # type: ignore
def reinitialize():
    global last_set_enable_jwt  # pylint: disable=global-variable-not-assigned
    global last_set_enable_anti_csrf  # pylint: disable=global-variable-not-assigned
    jwt_property_name: Union[str, None] = None
    json: Dict[str, Any] = request.get_json(silent=True)  # type: ignore
    if "jwtPropertyName" in json:
        jwt_property_name = json["jwtPropertyName"]

    Supertokens.reset()
    SessionRecipe.reset()
    config(last_set_enable_anti_csrf, last_set_enable_jwt, jwt_property_name)
    return "", 200


@app.route("/checkAllowCredentials", methods=["OPTIONS"])  # type: ignore
def check_allow_credentials_options():
    return send_options_api_response()


@app.route("/checkAllowCredentials", methods=["GET"])  # type: ignore
def check_allow_credentials():
    return jsonify(json.dumps("allow-credentials" in request.headers))  # type: ignore


@app.route("/testError", methods=["GET", "OPTIONS"])  # type: ignore
def test_error():
    if request.method == "OPTIONS":  # type: ignore
        return send_options_api_response()

    status_code = int(request.args.get("code", "500"))
    return Response("test error message", status=status_code)


if __name__ == "__main__":
    app.run(host="0.0.0.0", port=int(get_app_port()), threaded=True)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, List, Union

from supertokens_python.async_to_sync_wrapper import sync
from supertokens_python.recipe.passwordless.interfaces import (
    DeleteUserInfoOkResult,
    DeleteUserInfoUnknownUserIdError,
)

from .. import asyncio, interfaces
from ..types import EmailTemplateVars, SMSTemplateVars, User


def get_user_by_id(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[None, User]:
    from ..asyncio import get_user_by_id

    return sync(get_user_by_id(user_id, user_context))


def get_user_by_third_party_info(
    third_party_id: str,
    third_party_user_id: str,
    user_context: Union[None, Dict[str, Any]] = None,
):
    from ..asyncio import get_user_by_third_party_info

    return sync(
        get_user_by_third_party_info(third_party_id, third_party_user_id, user_context)
    )


def thirdparty_sign_in_up(
    third_party_id: str,
    third_party_user_id: str,
    email: str,
    user_context: Union[None, Dict[str, Any]] = None,
):
    from ..asyncio import thirdparty_sign_in_up

    return sync(
        thirdparty_sign_in_up(third_party_id, third_party_user_id, email, user_context)
    )


def get_users_by_email(
    email: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[User]:
    from ..asyncio import get_users_by_email

    return sync(get_users_by_email(email, user_context))


def create_code(
    email: Union[None, str] = None,
    phone_number: Union[None, str] = None,
    user_input_code: Union[None, str] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> interfaces.CreateCodeOkResult:
    return sync(
        asyncio.create_code(
            email=email,
            phone_number=phone_number,
            user_input_code=user_input_code,
            user_context=user_context,
        )
    )


def create_new_code_for_device(
    device_id: str,
    user_input_code: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[
    interfaces.CreateNewCodeForDeviceOkResult,
    interfaces.CreateNewCodeForDeviceRestartFlowError,
    interfaces.CreateNewCodeForDeviceUserInputCodeAlreadyUsedError,
]:
    return sync(
        asyncio.create_new_code_for_device(
            device_id=device_id,
            user_input_code=user_input_code,
            user_context=user_context,
        )
    )


def consume_code(
    pre_auth_session_id: str,
    user_input_code: Union[str, None] = None,
    device_id: Union[str, None] = None,
    link_code: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[
    interfaces.ConsumeCodeOkResult,
    interfaces.ConsumeCodeIncorrectUserInputCodeError,
    interfaces.ConsumeCodeExpiredUserInputCodeError,
    interfaces.ConsumeCodeRestartFlowError,
]:
    return sync(
        asyncio.consume_code(
            pre_auth_session_id=pre_auth_session_id,
            user_input_code=user_input_code,
            device_id=device_id,
            link_code=link_code,
            user_context=user_context,
        )
    )


def get_user_by_phone_number(
    phone_number: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[User, None]:
    return sync(
        asyncio.get_user_by_phone_number(
            phone_number=phone_number, user_context=user_context
        )
    )


def update_passwordless_user(
    user_id: str,
    email: Union[str, None] = None,
    phone_number: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[
    interfaces.PasswordlessUpdateUserOkResult,
    interfaces.PasswordlessUpdateUserUnknownUserIdError,
    interfaces.PasswordlessUpdateUserEmailAlreadyExistsError,
    interfaces.PasswordlessUpdateUserPhoneNumberAlreadyExistsError,
]:
    return sync(
        asyncio.update_passwordless_user(
            user_id=user_id,
            email=email,
            phone_number=phone_number,
            user_context=user_context,
        )
    )


def delete_email_for_passwordless_user(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[DeleteUserInfoOkResult, DeleteUserInfoUnknownUserIdError]:
    return sync(
        asyncio.delete_email_for_passwordless_user(
            user_id=user_id, user_context=user_context
        )
    )


def delete_phone_number_for_user(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[DeleteUserInfoOkResult, DeleteUserInfoUnknownUserIdError]:
    return sync(
        asyncio.delete_phone_number_for_user(user_id=user_id, user_context=user_context)
    )


def revoke_all_codes(
    email: Union[str, None] = None,
    phone_number: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> interfaces.RevokeAllCodesOkResult:
    return sync(
        asyncio.revoke_all_codes(
            email=email, phone_number=phone_number, user_context=user_context
        )
    )


def revoke_code(
    code_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> interfaces.RevokeCodeOkResult:
    return sync(asyncio.revoke_code(code_id=code_id, user_context=user_context))


def list_codes_by_email(
    email: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[interfaces.DeviceType]:
    return sync(asyncio.list_codes_by_email(email=email, user_context=user_context))


def list_codes_by_phone_number(
    phone_number: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[interfaces.DeviceType]:
    return sync(
        asyncio.list_codes_by_phone_number(
            phone_number=phone_number, user_context=user_context
        )
    )


def list_codes_by_device_id(
    device_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[interfaces.DeviceType, None]:
    return sync(
        asyncio.list_codes_by_device_id(device_id=device_id, user_context=user_context)
    )


def list_codes_by_pre_auth_session_id(
    pre_auth_session_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[interfaces.DeviceType, None]:
    return sync(
        asyncio.list_codes_by_pre_auth_session_id(
            pre_auth_session_id=pre_auth_session_id, user_context=user_context
        )
    )


def create_magic_link(
    email: Union[str, None],
    phone_number: Union[str, None],
    user_context: Union[None, Dict[str, Any]] = None,
) -> str:
    return sync(
        asyncio.create_magic_link(
            email=email, phone_number=phone_number, user_context=user_context
        )
    )


def passwordlessSigninup(
    email: Union[str, None],
    phone_number: Union[str, None],
    user_context: Union[None, Dict[str, Any]] = None,
) -> interfaces.ConsumeCodeOkResult:
    return sync(
        asyncio.passwordlessSigninup(
            email=email, phone_number=phone_number, user_context=user_context
        )
    )


def send_email(
    input_: EmailTemplateVars, user_context: Union[None, Dict[str, Any]] = None
):
    return sync(asyncio.send_email(input_, user_context))


def send_sms(input_: SMSTemplateVars, user_context: Union[None, Dict[str, Any]] = None):
    return sync(asyncio.send_sms(input_, user_context))



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Dict, List, TypeVar, Union

from supertokens_python.recipe.session.exceptions import (
    raise_invalid_claims_exception,
    raise_unauthorised_exception,
)
from supertokens_python.utils import get_timestamp_ms

from .cookie_and_header import (
    clear_session_response_mutator,
    front_token_response_mutator,
    token_response_mutator,
)
from .interfaces import SessionClaim, SessionClaimValidator, SessionContainer
from .utils import HUNDRED_YEARS_IN_MS

_T = TypeVar("_T")


class Session(SessionContainer):
    async def revoke_session(self, user_context: Union[Any, None] = None) -> None:
        if user_context is None:
            user_context = {}
        await self.recipe_implementation.revoke_session(
            self.session_handle, user_context
        )

        self.response_mutators.append(
            clear_session_response_mutator(
                self.config,
                self.transfer_method,
            )
        )

    async def get_session_data(
        self, user_context: Union[Dict[str, Any], None] = None
    ) -> Dict[str, Any]:
        if user_context is None:
            user_context = {}
        session_info = await self.recipe_implementation.get_session_information(
            self.session_handle, user_context
        )
        if session_info is None:
            raise_unauthorised_exception("Session does not exist anymore.")

        return session_info.session_data

    async def update_session_data(
        self,
        new_session_data: Dict[str, Any],
        user_context: Union[Dict[str, Any], None] = None,
    ) -> None:
        if user_context is None:
            user_context = {}
        updated = await self.recipe_implementation.update_session_data(
            self.session_handle, new_session_data, user_context
        )
        if not updated:
            raise_unauthorised_exception("Session does not exist anymore.")

    async def update_access_token_payload(
        self,
        new_access_token_payload: Dict[str, Any],
        user_context: Union[Dict[str, Any], None] = None,
    ) -> None:
        if user_context is None:
            user_context = {}

        result = await self.recipe_implementation.regenerate_access_token(
            self.access_token, new_access_token_payload, user_context
        )
        if result is None:
            raise_unauthorised_exception("Session does not exist anymore.")

        self.access_token_payload = result.session.user_data_in_jwt
        if result.access_token is not None:
            self.access_token = result.access_token.token

            self.response_mutators.append(
                front_token_response_mutator(
                    self.user_id,
                    result.access_token.expiry,
                    self.access_token_payload,
                )
            )
            # We set the expiration to 100 years, because we can't really access the expiration of the refresh token everywhere we are setting it.
            # This should be safe to do, since this is only the validity of the cookie (set here or on the frontend) but we check the expiration of the JWT anyway.
            # Even if the token is expired the presence of the token indicates that the user could have a valid refresh
            # Setting them to infinity would require special case handling on the frontend and just adding 100 years seems enough.
            self.response_mutators.append(
                token_response_mutator(
                    self.config,
                    "access",
                    result.access_token.token,
                    get_timestamp_ms() + HUNDRED_YEARS_IN_MS,
                    self.transfer_method,
                )
            )

    def get_user_id(self, user_context: Union[Dict[str, Any], None] = None) -> str:
        return self.user_id

    def get_access_token_payload(
        self, user_context: Union[Dict[str, Any], None] = None
    ) -> Dict[str, Any]:
        return self.access_token_payload

    def get_handle(self, user_context: Union[Dict[str, Any], None] = None) -> str:
        return self.session_handle

    def get_access_token(self, user_context: Union[Dict[str, Any], None] = None) -> str:
        return self.access_token

    async def get_time_created(
        self, user_context: Union[Dict[str, Any], None] = None
    ) -> int:
        if user_context is None:
            user_context = {}
        session_info = await self.recipe_implementation.get_session_information(
            self.session_handle, user_context
        )
        if session_info is None:
            raise_unauthorised_exception("Session does not exist anymore.")

        return session_info.time_created

    async def get_expiry(self, user_context: Union[Dict[str, Any], None] = None) -> int:
        if user_context is None:
            user_context = {}
        session_info = await self.recipe_implementation.get_session_information(
            self.session_handle, user_context
        )
        if session_info is None:
            raise_unauthorised_exception("Session does not exist anymore.")

        return session_info.expiry

    async def assert_claims(
        self,
        claim_validators: List[SessionClaimValidator],
        user_context: Union[Dict[str, Any], None] = None,
    ) -> None:
        if user_context is None:
            user_context = {}

        validate_claim_res = await self.recipe_implementation.validate_claims(
            self.get_user_id(user_context),
            self.get_access_token_payload(user_context),
            claim_validators,
            user_context,
        )

        if validate_claim_res.access_token_payload_update is not None:
            await self.merge_into_access_token_payload(
                validate_claim_res.access_token_payload_update, user_context
            )

        validation_errors = validate_claim_res.invalid_claims
        if len(validation_errors) > 0:
            raise_invalid_claims_exception("INVALID_CLAIMS", validation_errors)

    async def fetch_and_set_claim(
        self, claim: SessionClaim[Any], user_context: Union[Dict[str, Any], None] = None
    ) -> None:
        if user_context is None:
            user_context = {}

        update = await claim.build(self.get_user_id(), user_context)
        return await self.merge_into_access_token_payload(update, user_context)

    async def set_claim_value(
        self,
        claim: SessionClaim[_T],
        value: _T,
        user_context: Union[Dict[str, Any], None] = None,
    ) -> None:
        if user_context is None:
            user_context = {}

        update = claim.add_to_payload_({}, value, user_context)
        return await self.merge_into_access_token_payload(update, user_context)

    async def get_claim_value(
        self, claim: SessionClaim[_T], user_context: Union[Dict[str, Any], None] = None
    ) -> Union[_T, None]:
        if user_context is None:
            user_context = {}

        return claim.get_value_from_payload(
            self.get_access_token_payload(user_context), user_context
        )

    async def remove_claim(
        self, claim: SessionClaim[Any], user_context: Union[Dict[str, Any], None] = None
    ) -> None:
        if user_context is None:
            user_context = {}

        update = claim.remove_from_payload_by_merge_({}, user_context)
        return await self.merge_into_access_token_payload(update, user_context)

    async def merge_into_access_token_payload(
        self,
        access_token_payload_update: Dict[str, Any],
        user_context: Union[Dict[str, Any], None] = None,
    ) -> None:
        if user_context is None:
            user_context = {}

        update_payload = {
            **self.get_access_token_payload(user_context),
            **access_token_payload_update,
        }
        for k in access_token_payload_update.keys():
            if access_token_payload_update[k] is None:
                del update_payload[k]

        await self.update_access_token_payload(update_payload, user_context)



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from typing import Any, Dict, Union

from supertokens_python.async_to_sync_wrapper import sync

from ..interfaces import SignInOkResult, SignInWrongCredentialsError
from ..types import EmailTemplateVars, User


def update_email_or_password(
    user_id: str,
    email: Union[str, None] = None,
    password: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
):
    from supertokens_python.recipe.emailpassword.asyncio import update_email_or_password

    return sync(update_email_or_password(user_id, email, password, user_context))


def get_user_by_id(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[None, User]:
    from supertokens_python.recipe.emailpassword.asyncio import get_user_by_id

    return sync(get_user_by_id(user_id, user_context))


def get_user_by_email(
    email: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[None, User]:
    from supertokens_python.recipe.emailpassword.asyncio import get_user_by_email

    return sync(get_user_by_email(email, user_context))


def create_reset_password_token(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
):
    from supertokens_python.recipe.emailpassword.asyncio import (
        create_reset_password_token,
    )

    return sync(create_reset_password_token(user_id, user_context))


def reset_password_using_token(
    token: str, new_password: str, user_context: Union[None, Dict[str, Any]] = None
):
    from supertokens_python.recipe.emailpassword.asyncio import (
        reset_password_using_token,
    )

    return sync(reset_password_using_token(token, new_password, user_context))


def sign_in(
    email: str, password: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[SignInOkResult, SignInWrongCredentialsError]:
    from supertokens_python.recipe.emailpassword.asyncio import sign_in

    return sync(sign_in(email, password, user_context))


def sign_up(
    email: str, password: str, user_context: Union[None, Dict[str, Any]] = None
):
    from supertokens_python.recipe.emailpassword.asyncio import sign_up

    return sync(sign_up(email, password, user_context))


def send_email(
    input_: EmailTemplateVars, user_context: Union[None, Dict[str, Any]] = None
):
    from supertokens_python.recipe.emailpassword.asyncio import send_email

    return sync(send_email(input_, user_context))



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from typing import Any, Dict, List, Union

from supertokens_python.recipe.passwordless.interfaces import (
    DeleteUserInfoOkResult,
    DeleteUserInfoUnknownUserIdError,
)

from .. import interfaces
from ..recipe import ThirdPartyPasswordlessRecipe
from ..types import EmailTemplateVars, SMSTemplateVars, User


async def get_user_by_id(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[None, User]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.get_user_by_id(
        user_id, user_context
    )


async def get_user_by_third_party_info(
    third_party_id: str,
    third_party_user_id: str,
    user_context: Union[None, Dict[str, Any]] = None,
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.get_user_by_thirdparty_info(
        third_party_id, third_party_user_id, user_context
    )


async def thirdparty_sign_in_up(
    third_party_id: str,
    third_party_user_id: str,
    email: str,
    user_context: Union[None, Dict[str, Any]] = None,
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.thirdparty_sign_in_up(
        third_party_id, third_party_user_id, email, user_context
    )


async def get_users_by_email(
    email: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[User]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.get_users_by_email(
        email, user_context
    )


async def create_code(
    email: Union[None, str] = None,
    phone_number: Union[None, str] = None,
    user_input_code: Union[None, str] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> interfaces.CreateCodeOkResult:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.create_code(
        email=email,
        phone_number=phone_number,
        user_input_code=user_input_code,
        user_context=user_context,
    )


async def create_new_code_for_device(
    device_id: str,
    user_input_code: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[
    interfaces.CreateNewCodeForDeviceOkResult,
    interfaces.CreateNewCodeForDeviceRestartFlowError,
    interfaces.CreateNewCodeForDeviceUserInputCodeAlreadyUsedError,
]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.create_new_code_for_device(
        device_id=device_id, user_input_code=user_input_code, user_context=user_context
    )


async def consume_code(
    pre_auth_session_id: str,
    user_input_code: Union[str, None] = None,
    device_id: Union[str, None] = None,
    link_code: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[
    interfaces.ConsumeCodeOkResult,
    interfaces.ConsumeCodeIncorrectUserInputCodeError,
    interfaces.ConsumeCodeExpiredUserInputCodeError,
    interfaces.ConsumeCodeRestartFlowError,
]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.consume_code(
        pre_auth_session_id=pre_auth_session_id,
        user_input_code=user_input_code,
        device_id=device_id,
        link_code=link_code,
        user_context=user_context,
    )


async def get_user_by_phone_number(
    phone_number: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[User, None]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.get_user_by_phone_number(
        phone_number=phone_number, user_context=user_context
    )


async def update_passwordless_user(
    user_id: str,
    email: Union[str, None] = None,
    phone_number: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> Union[
    interfaces.PasswordlessUpdateUserOkResult,
    interfaces.PasswordlessUpdateUserUnknownUserIdError,
    interfaces.PasswordlessUpdateUserEmailAlreadyExistsError,
    interfaces.PasswordlessUpdateUserPhoneNumberAlreadyExistsError,
]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.update_passwordless_user(
        user_id=user_id,
        email=email,
        phone_number=phone_number,
        user_context=user_context,
    )


async def delete_email_for_passwordless_user(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[DeleteUserInfoOkResult, DeleteUserInfoUnknownUserIdError]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.delete_email_for_passwordless_user(
        user_id=user_id, user_context=user_context
    )


async def delete_phone_number_for_user(
    user_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[DeleteUserInfoOkResult, DeleteUserInfoUnknownUserIdError]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.delete_phone_number_for_user(
        user_id=user_id, user_context=user_context
    )


async def revoke_all_codes(
    email: Union[str, None] = None,
    phone_number: Union[str, None] = None,
    user_context: Union[None, Dict[str, Any]] = None,
) -> interfaces.RevokeAllCodesOkResult:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.revoke_all_codes(
        email=email, phone_number=phone_number, user_context=user_context
    )


async def revoke_code(
    code_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> interfaces.RevokeCodeOkResult:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.revoke_code(
        code_id=code_id, user_context=user_context
    )


async def list_codes_by_email(
    email: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[interfaces.DeviceType]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.list_codes_by_email(
        email=email, user_context=user_context
    )


async def list_codes_by_phone_number(
    phone_number: str, user_context: Union[None, Dict[str, Any]] = None
) -> List[interfaces.DeviceType]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.list_codes_by_phone_number(
        phone_number=phone_number, user_context=user_context
    )


async def list_codes_by_device_id(
    device_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[interfaces.DeviceType, None]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.list_codes_by_device_id(
        device_id=device_id, user_context=user_context
    )


async def list_codes_by_pre_auth_session_id(
    pre_auth_session_id: str, user_context: Union[None, Dict[str, Any]] = None
) -> Union[interfaces.DeviceType, None]:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().recipe_implementation.list_codes_by_pre_auth_session_id(
        pre_auth_session_id=pre_auth_session_id, user_context=user_context
    )


async def create_magic_link(
    email: Union[str, None],
    phone_number: Union[str, None],
    user_context: Union[None, Dict[str, Any]] = None,
) -> str:
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().passwordless_recipe.create_magic_link(
        email=email, phone_number=phone_number, user_context=user_context
    )


async def passwordlessSigninup(
    email: Union[str, None],
    phone_number: Union[str, None],
    user_context: Union[None, Dict[str, Any]] = None,
) -> interfaces.ConsumeCodeOkResult:
    if user_context is None:
        user_context = {}
    result = (
        await ThirdPartyPasswordlessRecipe.get_instance().passwordless_recipe.signinup(
            email=email, phone_number=phone_number, user_context=user_context
        )
    )
    return interfaces.ConsumeCodeOkResult(
        result.created_new_user,
        User(
            result.user.user_id,
            result.user.email,
            result.user.phone_number,
            None,
            result.user.time_joined,
        ),
    )


async def send_email(
    input_: EmailTemplateVars, user_context: Union[None, Dict[str, Any]] = None
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().email_delivery.ingredient_interface_impl.send_email(
        input_, user_context
    )


async def send_sms(
    input_: SMSTemplateVars, user_context: Union[None, Dict[str, Any]] = None
):
    if user_context is None:
        user_context = {}
    return await ThirdPartyPasswordlessRecipe.get_instance().sms_delivery.ingredient_interface_impl.send_sms(
        input_, user_context
    )



# Copyright (c) 2021, VRAI Labs and/or its affiliates. All rights reserved.
#
# This software is licensed under the Apache License, Version 2.0 (the
# "License") as published by the Apache Software Foundation.
#
# You may not use this file except in compliance with the License. You may
# obtain a copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from typing import Any, Dict, Callable, Union, TypeVar

from supertokens_python.ingredients.smsdelivery.services.twilio import (
    normalize_twilio_settings,
)
from supertokens_python.ingredients.smsdelivery.types import (
    SMSDeliveryInterface,
    TwilioServiceInterface,
    TwilioSettings,
)
from supertokens_python.recipe.passwordless.types import (
    PasswordlessLoginSMSTemplateVars,
)

from twilio.rest import Client  # type: ignore

from .service_implementation import ServiceImplementation

_T = TypeVar("_T")


class TwilioService(SMSDeliveryInterface[PasswordlessLoginSMSTemplateVars]):
    service_implementation: TwilioServiceInterface[PasswordlessLoginSMSTemplateVars]

    def __init__(
        self,
        twilio_settings: TwilioSettings,
        override: Union[
            Callable[[TwilioServiceInterface[_T]], TwilioServiceInterface[_T]], None
        ] = None,
    ) -> None:
        self.config = normalize_twilio_settings(twilio_settings)
        otps = twilio_settings.opts if twilio_settings.opts else {}
        self.twilio_client = Client(  # type: ignore
            twilio_settings.account_sid, twilio_settings.auth_token, **otps
        )
        oi = ServiceImplementation(self.twilio_client)  # type: ignore
        self.service_implementation = oi if override is None else override(oi)

    async def send_sms(
        self,
        template_vars: PasswordlessLoginSMSTemplateVars,
        user_context: Dict[str, Any],
    ) -> None:
        content = await self.service_implementation.get_content(
            template_vars, user_context
        )
        await self.service_implementation.send_raw_sms(
            content,
            user_context,
            from_=self.config.from_,
            messaging_service_sid=self.config.messaging_service_sid,
        )
